;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Emacs pov-mode supplementary file
;;
;; This file implements the insert menu like known
;; from the Windows version of POV-Ray
;;
;; This file is built from the data basis of official POV-Ray 3.5
;; on Nov 22 2002 by Christoph Hormann <chris_hormann@gmx.de>
;;
;; NOTE: This is surely not the most elegant method
;;       to implement the menu, but it works.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;;  Installation instructions:
;;    - add the following to the end of your pov-mode.el
;;      file:
;;
;;      (require 'pov-im)
;;      (pov-im-generate pov-mode-map)
;;
;;    - add the following to the 'pov-mode' function in
;;      your pov-mode.el file:
;;
;;      (easy-menu-add pov-im-menu)
;;
;;    pov-mode 2.8 or newer already contain these lines, those
;;    at the end of the file are just commented out by default.
;;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(require 'easymenu)
(provide 'pov-im)

(defvar pov-im-menu nil)


(defun pov-im-generate (mode-map)
  (easy-menu-define
    pov-im-menu
    mode-map
    "The POV-Ray Insert menu."
    '("Insert"
               ("Scene templates"
                 ["Basic scene" pov-im-10---Basic-scene t]
                 ["Orthographic scene" pov-im-20---Orthographic-scene t]
                 ["Radiosity scene" pov-im-30---Radiosity-scene t]
                 ["Photons scene" pov-im-40---Photons-scene t]
                 ["Checkered floor" pov-im-50---Checkered-floor t]
                 ["Image Map" pov-im-60---Image-Map t]
                 ["Text Fonts" pov-im-70---Text-Fonts t]
               )
               ("Headers"
                 ["Scene File Header" pov-im-10---Scene-File-Header t]
                 ["Standard includes" pov-im-20---Standard-includes t]
                 ["Additional includes" pov-im-30---Additional-includes t]
               )
               ("Include files"
                 ["'arrays.inc'" pov-im--arrays.inc- t]
                 ["'colors.inc'" pov-im--colors.inc- t]
                 ["'consts.inc'" pov-im--consts.inc- t]
                 ["'finish.inc'" pov-im--finish.inc- t]
                 ["'functions.inc'" pov-im--functions.inc- t]
                 ["'glass.inc'" pov-im--glass.inc- t]
                 ["'math.inc'" pov-im--math.inc- t]
                 ["'metals.inc'" pov-im--metals.inc- t]
                 ["'rand.inc'" pov-im--rand.inc- t]
                 ["'shapes.inc'" pov-im--shapes.inc- t]
                 ["'shapes2.inc'" pov-im--shapes2.inc- t]
                 ["'skies.inc'" pov-im--skies.inc- t]
                 ["'stones.inc'" pov-im--stones.inc- t]
                 ["'strings.inc'" pov-im--strings.inc- t]
                 ["'transforms.inc'" pov-im--transforms.inc- t]
                 ["'woods.inc'" pov-im--woods.inc- t]
               )
               ("Statements"
                 ["background" pov-im-10---background t]
                 ["global_settings" pov-im-20---global_settings t]
                 "----"
                 ["adc_bailout" pov-im-30---adc_bailout t]
                 ["ambient_light" pov-im-40---ambient_light t]
                 ["assumed_gamma" pov-im-50---assumed_gamma t]
                 ["charset" pov-im-60---charset t]
                 ["hf_gray_16" pov-im-70---hf_gray_16 t]
                 ["irid_wavelength" pov-im-80---irid_wavelength t]
                 ["max_intersections" pov-im-90---max_intersections t]
                 ["max_trace_level" pov-im-A0---max_trace_level t]
                 ["noise_generator" pov-im-B0---noise_generator t]
                 ["number_of_waves" pov-im-C0---number_of_waves t]
               )
               ("Expressions"
                 ["Bounds extents" pov-im-Bounds-extents t]
                 ["Built-in Ids" pov-im-Built-in-Ids t]
                 ["Conditional Expression" pov-im-Conditional-Expression t]
                 ["Float Functions" pov-im-Float-Functions t]
                 ["Logical Operators" pov-im-Logical-Operators t]
                 ["Math Operators" pov-im-Math-Operators t]
                 ["Relational Operators" pov-im-Relational-Operators t]
                 ["Strings" pov-im-Strings t]
                 ["Vector Functions" pov-im-Vector-Functions t]
                 ["Vector components" pov-im-Vector-components t]
                 ["inside" pov-im-inside t]
                 ["rand" pov-im-rand t]
                 ["trace" pov-im-trace t]
               )
               ("Misc. Directives"
                 ["declare" pov-im-declare t]
                 ["default" pov-im-default t]
                 ["if example" pov-im-if-example t]
                 ["if" pov-im-if t]
                 ["ifdef example" pov-im-ifdef-example t]
                 ["ifdef" pov-im-ifdef t]
                 ["local" pov-im-local t]
                 ["macro" pov-im-macro t]
                 ["messages" pov-im-messages t]
                 ["switch example" pov-im-switch-example t]
                 ["switch" pov-im-switch t]
                 ["undef" pov-im-undef t]
                 ["version" pov-im-version t]
                 ["while" pov-im-while t]
               )
               ("Animation"
                 ["clock" pov-im-10---clock t]
                 ["additional animation variables" pov-im-20---additional-animation-variables t]
               )
               ("Colors"
                 ["rgb" pov-im-10---rgb t]
                 ["rgbf" pov-im-20---rgbf t]
                 ["rgbt" pov-im-30---rgbt t]
                 ["rgbft" pov-im-40---rgbft t]
                 "----"
                 ["red" pov-im-50---red t]
                 ["green" pov-im-60---green t]
                 ["blue" pov-im-70---blue t]
                 ["filter" pov-im-80---filter t]
                 ["transmit" pov-im-90---transmit t]
                 "----"
                 ["color components" pov-im-A0---color-components t]
               )
               ("Special effects"
                 ["Fog definition" pov-im-10---Fog-definition t]
                 ["Ambient fog" pov-im-20---Ambient-fog t]
                 ["Ground fog" pov-im-30---Ground-fog t]
                 ["Atmospheric media" pov-im-40---Atmospheric-media t]
                 "----"
                 ["sky_sphere" pov-im-50---sky_sphere t]
                 ["rainbow" pov-im-60---rainbow t]
                 "----"
                 ["Global photons" pov-im-70---Global-photons t]
                 ["Light source photons" pov-im-80---Light-source-photons t]
                 ["Object photons" pov-im-90---Object-photons t]
                 "----"
                 ["radiosity" pov-im-A0---radiosity t]
               )
               ("Cameras"
                 ["A typical camera" pov-im-A-typical-camera t]
                 ["Camera definition" pov-im-Camera-definition t]
                 ["cylinder camera" pov-im-cylinder-camera t]
                 ["fisheye camera" pov-im-fisheye-camera t]
                 ["focal blur camera" pov-im-focal-blur-camera t]
                 ["normal perturbed camera" pov-im-normal-perturbed-camera t]
                 ["omnimax camera" pov-im-omnimax-camera t]
                 ["orthographic camera" pov-im-orthographic-camera t]
                 ["panoramic camera" pov-im-panoramic-camera t]
                 ["perspective camera" pov-im-perspective-camera t]
                 ["sky camera" pov-im-sky-camera t]
                 ["spherical camera" pov-im-spherical-camera t]
                 ["ultra_wide_angle camera" pov-im-ultra_wide_angle-camera t]
               )
               ("Light sources"
                 ["Light definition" pov-im-Light-definition t]
                 ["area_light" pov-im-area_light t]
                 ["cylindrical light" pov-im-cylindrical-light t]
                 ["light fading" pov-im-light-fading t]
                 ["light_group" pov-im-light_group t]
                 ["looks_like" pov-im-looks_like t]
                 ["parallel" pov-im-parallel t]
                 ["point light" pov-im-point-light t]
                 ["projected_through" pov-im-projected_through t]
                 ["spotlight" pov-im-spotlight t]
               )
               ("Shapes"
                 ["bicubic_patch" pov-im-bicubic_patch t]
                 ["blob" pov-im-blob t]
                 ["box" pov-im-box t]
                 ["cone" pov-im-cone t]
                 ["cubic" pov-im-cubic t]
                 ["cylinder" pov-im-cylinder t]
                 ["disc" pov-im-disc t]
                 ["fractal" pov-im-fractal t]
                 ["height_field+imagemap" pov-im-height_field+imagemap t]
                 ["height_field" pov-im-height_field t]
                 ["isosurface" pov-im-isosurface t]
                 ["lathe" pov-im-lathe t]
                 ["mesh" pov-im-mesh t]
                 ["mesh2" pov-im-mesh2 t]
                 ["object" pov-im-object t]
                 ["parametric" pov-im-parametric t]
                 ["plane" pov-im-plane t]
                 ["polygon" pov-im-polygon t]
                 ["polynomial" pov-im-polynomial t]
                 ["prism" pov-im-prism t]
                 ["quadric" pov-im-quadric t]
                 ["quartic" pov-im-quartic t]
                 ["smooth_triangle" pov-im-smooth_triangle t]
                 ["sphere" pov-im-sphere t]
                 ["sphere_sweep" pov-im-sphere_sweep t]
                 ["superellipsoid" pov-im-superellipsoid t]
                 ["surface of revolution" pov-im-surface-of-revolution t]
                 ["text" pov-im-text t]
                 ["torus" pov-im-torus t]
                 ["triangle" pov-im-triangle t]
               )
               ("Shapes (shorter versions)"
                 ["box" pov-im-box t]
                 ["cone" pov-im-cone t]
                 ["cylinder" pov-im-cylinder t]
                 ["disc" pov-im-disc t]
                 ["plane" pov-im-plane t]
                 ["sphere" pov-im-sphere t]
                 ["superellipsoid" pov-im-superellipsoid t]
                 ["torus" pov-im-torus t]
               )
               ("Shape modifiers"
                 ["bounded_by" pov-im-bounded_by t]
                 ["clipped_by" pov-im-clipped_by t]
                 ["double_illuminate" pov-im-double_illuminate t]
                 ["hollow" pov-im-hollow t]
                 ["no_image" pov-im-no_image t]
                 ["no_reflection" pov-im-no_reflection t]
                 ["no_shadow" pov-im-no_shadow t]
                 ["open" pov-im-open t]
                 ["sturm" pov-im-sturm t]
               )
               ("CSG operations"
                 ["cutaway_textures" pov-im-cutaway_textures t]
                 ["difference" pov-im-difference t]
                 ["intersection" pov-im-intersection t]
                 ["inverse" pov-im-inverse t]
                 ["merge" pov-im-merge t]
                 ["union" pov-im-union t]
               )
               ("Material"
                 ["interior" pov-im-interior t]
                 ["material" pov-im-material t]
                 ["media" pov-im-media t]
               )
               ("Textures"
                 ("Texture Maps"
                   ["bump_map normal" pov-im-bump_map-normal t]
                   ["image_map pigment" pov-im-image_map-pigment t]
                   ["material_map texture" pov-im-material_map-texture t]
                   ["tiles texture" pov-im-tiles-texture t]
                 )
                 ["color" pov-im-color t]
                 ["color_map" pov-im-color_map t]
                 ["finish properties" pov-im-finish-properties t]
                 ["function image type" pov-im-function-image-type t]
                 ["interior_texture" pov-im-interior_texture t]
                 ["normal component" pov-im-normal-component t]
                 ["pigment component" pov-im-pigment-component t]
                 ["quick_color attribute" pov-im-quick_color-attribute t]
                 ["texture example" pov-im-texture-example t]
                 ["triangle texture interpolation" pov-im-triangle-texture-interpolation t]
                 ["uv_mapping" pov-im-uv_mapping t]
               )
               ("Patterns 1 (a-m)"
                 ["agate" pov-im-agate t]
                 ["average" pov-im-average t]
                 ["boxed" pov-im-boxed t]
                 ["bozo" pov-im-bozo t]
                 ["brick" pov-im-brick t]
                 ["bumps" pov-im-bumps t]
                 ["cells" pov-im-cells t]
                 ["checker" pov-im-checker t]
                 ["crackle" pov-im-crackle t]
                 ["cylindrical" pov-im-cylindrical t]
                 ["density_file" pov-im-density_file t]
                 ["dents" pov-im-dents t]
                 ["facets" pov-im-facets t]
                 ["fractal" pov-im-fractal t]
                 ["function" pov-im-function t]
                 ["gradient" pov-im-gradient t]
                 ["granite" pov-im-granite t]
                 ["hexagon" pov-im-hexagon t]
                 ["image_pattern" pov-im-image_pattern t]
                 ["leopard" pov-im-leopard t]
                 ["marble" pov-im-marble t]
               )
               ("Patterns 2 (n-z)"
                 ["object" pov-im-object t]
                 ["onion" pov-im-onion t]
                 ["pigment_pattern" pov-im-pigment_pattern t]
                 ["planar" pov-im-planar t]
                 ["quilted" pov-im-quilted t]
                 ["radial" pov-im-radial t]
                 ["ripples" pov-im-ripples t]
                 ["slope" pov-im-slope t]
                 ["spherical" pov-im-spherical t]
                 ["spiral1" pov-im-spiral1 t]
                 ["spiral2" pov-im-spiral2 t]
                 ["spotted" pov-im-spotted t]
                 ["waves" pov-im-waves t]
                 ["wood" pov-im-wood t]
                 ["wrinkles" pov-im-wrinkles t]
               )
               ("Pattern-Texture attributes"
                 ["waveform attribute" pov-im-10---waveform-attribute t]
                 ["frequency attribute" pov-im-20---frequency-attribute t]
                 ["phase attribute" pov-im-30---phase-attribute t]
                 ["turbulence attribute" pov-im-40---turbulence-attribute t]
                 "----"
                 ["turbulence warp" pov-im-50---turbulence-warp t]
                 ["black_hole warp" pov-im-60---black_hole-warp t]
                 ["repeat warp" pov-im-70---repeat-warp t]
                 ["mapping warps" pov-im-80---mapping-warps t]
                 "----"
                 ["noise_generator" pov-im-90---noise_generator t]
               )
               ("Transformations"
                 ["matrix" pov-im-matrix t]
                 ["rotate" pov-im-rotate t]
                 ["scale" pov-im-scale t]
                 ["transform" pov-im-transform t]
                 ["translate" pov-im-translate t]
               )
    )
  )
)

(defun pov-im-10---Basic-scene ()
  (interactive
    (insert (concat
                    "// Persistence of Vision Ray Tracer Scene Description File\n"
                    "// File: ?.pov\n"
                    "// Vers: 3.5\n"
                    "// Desc: Basic Scene Example\n"
                    "// Date: mm/dd/yy\n"
                    "// Auth: ?\n"
                    "//\n"
                    "\n"
                    "#version 3.5;\n"
                    "\n"
                    "#include \"colors.inc\"\n"
                    "\n"
                    "global_settings {\n"
                    "  assumed_gamma 1.0\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "camera {\n"
                    "  location  <0.0, 0.5, -4.0>\n"
                    "  direction 1.5*z\n"
                    "  right     x*image_width/image_height\n"
                    "  look_at   <0.0, 0.0,  0.0>\n"
                    "}\n"
                    "\n"
                    "sky_sphere {\n"
                    "  pigment {\n"
                    "    gradient y\n"
                    "    color_map {\n"
                    "      [0.0 rgb <0.6,0.7,1.0>]\n"
                    "      [0.7 rgb <0.0,0.1,0.8>]\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "light_source {\n"
                    "  <0, 0, 0>            // light's position (translated below)\n"
                    "  color rgb <1, 1, 1>  // light's color\n"
                    "  translate <-30, 30, -30>\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "plane {\n"
                    "  y, -1\n"
                    "  pigment { color rgb <0.7,0.5,0.3> }\n"
                    "}\n"
                    "\n"
                    "sphere {\n"
                    "  0.0, 1\n"
                    "  texture {\n"
                    "    pigment {\n"
                    "      radial\n"
                    "      frequency 8\n"
                    "      color_map {\n"
                    "        [0.00 color rgb <1.0,0.4,0.2> ]\n"
                    "        [0.33 color rgb <0.2,0.4,1.0> ]\n"
                    "        [0.66 color rgb <0.4,1.0,0.2> ]\n"
                    "        [1.00 color rgb <1.0,0.4,0.2> ]\n"
                    "      }\n"
                    "    }\n"
                    "    finish{\n"
                    "      specular 0.6\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-20---Orthographic-scene ()
  (interactive
    (insert (concat
                    "// Persistence of Vision Ray Tracer Scene Description File\n"
                    "// File: ?.pov\n"
                    "// Vers: 3.5\n"
                    "// Desc: Orthographic Scene Example\n"
                    "//       useful for generating image_maps, heightfields, etc.\n"
                    "// Date: mm/dd/yy\n"
                    "// Auth: ?\n"
                    "//\n"
                    "\n"
                    "#version 3.5;\n"
                    "\n"
                    "global_settings {\n"
                    "  assumed_gamma 1.0\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "camera {\n"
                    "  orthographic\n"
                    "  location <0,0,1>     // position & direction of view\n"
                    "  look_at  <0,0,0>\n"
                    "  right 1*x            // horizontal size of view  \___ to be rendered at square size\n"
                    "  up 1*y               // vertical size of view    /\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "box {                  // this box fits exactly in view\n"
                    "  <-0.5, -0.5, 0>, <0.5, 0.5, 0>\n"
                    "  texture {\n"
                    "    pigment {\n"
                    "      agate\n"
                    "      color_map {\n"
                    "        [0.0 color rgb 0.0 ]\n"
                    "        [1.0 color rgb 1.0 ]\n"
                    "      }\n"
                    "    }\n"
                    "    finish {\n"
                    "      ambient 1.0\n"
                    "      diffuse 0.0\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-30---Radiosity-scene ()
  (interactive
    (insert (concat
                    "// Persistence of Vision Ray Tracer Scene Description File\n"
                    "// File: ?.pov\n"
                    "// Vers: 3.5\n"
                    "// Desc: Radiosity Scene Template\n"
                    "// Date: mm/dd/yy\n"
                    "// Auth: ?\n"
                    "//\n"
                    "\n"
                    "#version 3.5;\n"
                    "\n"
                    "#declare Radiosity=on;\n"
                    "\n"
                    "global_settings {\n"
                    "  assumed_gamma 1.0\n"
                    "  //max_trace_level 25\n"
                    "  #if (Radiosity)\n"
                    "    radiosity {\n"
                    "      pretrace_start 0.08           // start pretrace at this size\n"
                    "      pretrace_end   0.04           // end pretrace at this size\n"
                    "      count 35                      // higher -> higher quality (1..1600) [35]\n"
                    "      nearest_count 5               // higher -> higher quality (1..10) [5]\n"
                    "      error_bound 1.8               // higher -> smoother, less accurate [1.8]\n"
                    "      recursion_limit 3             // how much interreflections are calculated (1..5+) [3]\n"
                    "      low_error_factor .5           // reduce error_bound during last pretrace step\n"
                    "      gray_threshold 0.0            // increase for weakening colors (0..1) [0]\n"
                    "      minimum_reuse 0.015           // reuse of old radiosity samples [0.015]\n"
                    "      brightness 1                  // brightness of radiosity effects (0..1) [1]\n"
                    "\n"
                    "      adc_bailout 0.01/2\n"
                    "      //normal on                   // take surface normals into account [off]\n"
                    "      //media on                    // take media into account [off]\n"
                    "      //save_file \"file_name\"       // save radiosity data\n"
                    "      //load_file \"file_name\"       // load saved radiosity data\n"
                    "      //always_sample off           // turn sampling in final trace off [on]\n"
                    "      //max_sample 1.0              // maximum brightness of samples\n"
                    "    }\n"
                    "  #end\n"
                    "}\n"
                    "\n"
                    "#default {\n"
                    "  texture {\n"
                    "    pigment {rgb 1}\n"
                    "    #if (Radiosity)\n"
                    "      finish {\n"
                    "        ambient 0.0\n"
                    "        diffuse 0.6\n"
                    "        specular 0.3\n"
                    "      }\n"
                    "    #else\n"
                    "      finish {\n"
                    "        ambient 0.1\n"
                    "        diffuse 0.6\n"
                    "        specular 0.3\n"
                    "      }\n"
                    "    #end\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "camera {\n"
                    "  right x*image_width/image_height\n"
                    "  location  <0,1.5,-4>\n"
                    "  look_at   <0,1,0>\n"
                    "}\n"
                    "\n"
                    "light_source {\n"
                    "  <500,500,-500>       // light's position\n"
                    "  color rgb <1, 1, 1>  // light's color\n"
                    "}\n"
                    "\n"
                    "sky_sphere {\n"
                    "  pigment {\n"
                    "    gradient y\n"
                    "    color_map {\n"
                    "      [0.0 rgb <0.6,0.7,1.0>]\n"
                    "      [0.7 rgb <0.0,0.1,0.8>]\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "plane {\n"
                    "  y, 0\n"
                    "  texture {\n"
                    "    pigment {\n"
                    "      checker\n"
                    "      color rgb <1.0, 0.8, 0.6>\n"
                    "      color rgb <1.0, 0.0, 0.0>\n"
                    "      scale 0.5\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "sphere {\n"
                    "  <0,1,0>, 1\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-40---Photons-scene ()
  (interactive
    (insert (concat
                    "// Persistence of Vision Ray Tracer Scene Description File\n"
                    "// File: ?.pov\n"
                    "// Vers: 3.5\n"
                    "// Desc: Photon Scene Template\n"
                    "// Date: mm/dd/yy\n"
                    "// Auth: ?\n"
                    "//\n"
                    "\n"
                    "#version 3.5;\n"
                    "\n"
                    "#declare Photons=on;\n"
                    "\n"
                    "global_settings {\n"
                    "  assumed_gamma 1.0\n"
                    "  max_trace_level 5\n"
                    "  #if (Photons)          // global photon block\n"
                    "    photons {\n"
                    "      spacing 0.02                 // specify the density of photons\n"
                    "      //count 100000               // alternatively use a total number of photons\n"
                    "\n"
                    "      //gather min, max            // amount of photons gathered during render [20, 100]\n"
                    "      //media max_steps [,factor]  // media photons\n"
                    "      //jitter 1.0                 // jitter phor photon rays\n"
                    "      //max_trace_level 5          // optional separate max_trace_level\n"
                    "      //adc_bailout 1/255          // see global adc_bailout\n"
                    "      //save_file \"filename\"       // save photons to file\n"
                    "      //load_file \"filename\"       // load photons from file\n"
                    "      //autostop 0                 // photon autostop option\n"
                    "      //radius 10                  // manually specified search radius\n"
                    "      // (---Adaptive Search Radius---)\n"
                    "      //steps 1\n"
                    "      //expand_thresholds 0.2, 40\n"
                    "    }\n"
                    "\n"
                    "  #end\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "camera {\n"
                    "  right x*image_width/image_height\n"
                    "  location  <0,1.6,-5>\n"
                    "  look_at   <0,0.75,0>\n"
                    "}\n"
                    "\n"
                    "light_source {\n"
                    "  <500,500,150>       // light's position\n"
                    "  color rgb 1.3       // light's color\n"
                    "  photons {           // photon block for a light source\n"
                    "    refraction on\n"
                    "    reflection on\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "sky_sphere {\n"
                    "  pigment {\n"
                    "    gradient y\n"
                    "    color_map {\n"
                    "      [0.0 rgb <0.6,0.7,1.0>]\n"
                    "      [0.7 rgb <0.0,0.1,0.8>]\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "plane {\n"
                    "  y, 0\n"
                    "  texture {\n"
                    "    pigment { color rgb <1.0, 0.8, 0.6> }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "\n"
                    "#declare M_Glass=    // Glass material\n"
                    "material {\n"
                    "  texture {\n"
                    "    pigment {rgbt 1}\n"
                    "    finish {\n"
                    "      ambient 0.0\n"
                    "      diffuse 0.05\n"
                    "      specular 0.6\n"
                    "      roughness 0.005\n"
                    "      reflection {\n"
                    "        0.1, 1.0\n"
                    "        fresnel on\n"
                    "      }\n"
                    "      conserve_energy\n"
                    "    }\n"
                    "  }\n"
                    "  interior {\n"
                    "    ior 1.5\n"
                    "    fade_power 1001\n"
                    "    fade_distance 0.9\n"
                    "    fade_color <0.5,0.8,0.6>\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "\n"
                    "sphere {\n"
                    "  <0,1,0>, 1\n"
                    "  translate <1.0,0,-1.3>\n"
                    "  material { M_Glass }\n"
                    "\n"
                    "  photons {  // photon block for an object\n"
                    "    target 1.0\n"
                    "    refraction on\n"
                    "    reflection on\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "cylinder {\n"
                    "  <0,0.01,0>, <0,2.5,0>, 1\n"
                    "  translate <-1.2,0,0.8>\n"
                    "  material { M_Glass }\n"
                    "\n"
                    "  photons {  // photon block for an object\n"
                    "    target 1.0\n"
                    "    refraction on\n"
                    "    reflection on\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-50---Checkered-floor ()
  (interactive
    (insert (concat
                    "// Persistence of Vision Ray Tracer Scene Description File\n"
                    "// File: ?.pov\n"
                    "// Vers: 3.5\n"
                    "// Desc: Checkered Floor Example\n"
                    "// Date: mm/dd/yy\n"
                    "// Auth: ?\n"
                    "//\n"
                    "\n"
                    "#version 3.5;\n"
                    "\n"
                    "#include \"colors.inc\"\n"
                    "\n"
                    "global_settings {\n"
                    "  assumed_gamma 1.0\n"
                    "  max_trace_level 5\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "camera {\n"
                    "  location  <0.0, 0.5, -4.0>\n"
                    "  direction 1.5*z\n"
                    "  right     x*image_width/image_height\n"
                    "  look_at   <0.0, 0.0,  0.0>\n"
                    "}\n"
                    "\n"
                    "sky_sphere {\n"
                    "  pigment {\n"
                    "    gradient y\n"
                    "    color_map {\n"
                    "      [0.0 rgb <0.6,0.7,1.0>]\n"
                    "      [0.7 rgb <0.0,0.1,0.8>]\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "light_source {\n"
                    "  <0, 0, 0>            // light's position (translated below)\n"
                    "  color rgb <1, 1, 1>  // light's color\n"
                    "  translate <-30, 30, -30>\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "plane {               // checkered floor\n"
                    "  y, -1\n"
                    "  texture\n"
                    "  {\n"
                    "    pigment {\n"
                    "      checker\n"
                    "      color rgb 1\n"
                    "      color blue 1\n"
                    "      scale 0.5\n"
                    "    }\n"
                    "    finish{\n"
                    "      diffuse 0.8\n"
                    "      ambient 0.1\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "sphere {              // reflective sphere\n"
                    "  0.0, 1\n"
                    "  texture {\n"
                    "    pigment {\n"
                    "      color rgb <0.8,0.8,1.0>\n"
                    "    }\n"
                    "    finish{\n"
                    "      diffuse 0.3\n"
                    "      ambient 0.0\n"
                    "      specular 0.6\n"
                    "      reflection {\n"
                    "        0.8\n"
                    "        metallic\n"
                    "      }\n"
                    "      conserve_energy\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-60---Image-Map ()
  (interactive
    (insert (concat
                    "// Persistence of Vision Ray Tracer Scene Description File\n"
                    "// File: ?.pov\n"
                    "// Vers: 3.5\n"
                    "// Desc: Image Map Example\n"
                    "// Date: mm/dd/yy\n"
                    "// Auth: ?\n"
                    "//\n"
                    "\n"
                    "#version 3.5;\n"
                    "\n"
                    "#include \"colors.inc\"\n"
                    "\n"
                    "global_settings {\n"
                    "  assumed_gamma 1.0\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "camera {\n"
                    "  location  <0.0, 0.0, -4.0>\n"
                    "  direction 2*z\n"
                    "  right     x*image_width/image_height\n"
                    "  look_at   <0.0, 0.0,  0.0>\n"
                    "}\n"
                    "\n"
                    "sky_sphere {\n"
                    "  pigment {\n"
                    "    gradient y\n"
                    "    color_map {\n"
                    "      [0.0 color blue 0.6]\n"
                    "      [1.0 color rgb 1]\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "light_source {\n"
                    "  <0, 0, 0>            // light's position (translated below)\n"
                    "  color rgb <1, 1, 1>  // light's color\n"
                    "  translate <-30, 30, -30>\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "plane {\n"
                    "  y, -1\n"
                    "  texture {\n"
                    "    pigment { checker color rgb 1 color blue 1 scale 0.5 }\n"
                    "    finish { reflection 0.2 }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "plane {\n"
                    "  z, -1\n"
                    "  texture {\n"
                    "    pigment {\n"
                    "      image_map {\n"
                    "        png \"test.png\"\n"
                    "        interpolate 2 // smooth it\n"
                    "        once   // don't tile image, just one copy\n"
                    "        filter 0 0.8  // make 1st color mostly transparent\n"
                    "        filter 1 0.8  // make 2nd color mostly transparent\n"
                    "      }\n"
                    "      // transform it to unit-size (-1 to +1)\n"
                    "      translate -0.5*(x+y) // center on the origin\n"
                    "      scale 2              // make it unit-sized\n"
                    "    }\n"
                    "    finish { ambient 0.3 }\n"
                    "  }\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-70---Text-Fonts ()
  (interactive
    (insert (concat
                    "// Persistence of Vision Ray Tracer Scene Description File\n"
                    "// File: ?.pov\n"
                    "// Vers: 3.5\n"
                    "// Desc: Basic TTF font Example\n"
                    "// Date: mm/dd/yy\n"
                    "// Auth: ?\n"
                    "//\n"
                    "\n"
                    "#version 3.5;\n"
                    "\n"
                    "global_settings {\n"
                    "  assumed_gamma 1.0\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "camera {\n"
                    "  location  <0.0, 2.0, -6.0>\n"
                    "  direction 1.5*z\n"
                    "  right     4/3*x\n"
                    "  look_at   <0.0, 0.0,  0.0>\n"
                    "}\n"
                    "\n"
                    "sky_sphere {\n"
                    "  pigment {\n"
                    "    gradient y\n"
                    "    color_map {\n"
                    "      [0.0 color blue 0.6]\n"
                    "      [1.0 color rgb 1]\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
                    "\n"
                    "light_source {\n"
                    "  <0, 0, 0>            // light's position (translated below)\n"
                    "  color rgb <1, 1, 1>  // light's color\n"
                    "  translate <-30, 30, -30>\n"
                    "}\n"
                    "\n"
                    "// ----------------------------------------\n"
                    "\n"
                    "#declare Text_Tex = texture {\n"
                    "  pigment { granite scale 0.5 }\n"
                    "  finish { specular 0.7 }\n"
                    "}\n"
                    "\n"
                    "text {\n"
                    "  ttf \"crystal.ttf\", \"Hello\",\n"
                    "  2, // depth\n"
                    "  0  // spacing\n"
                    "  texture { Text_Tex }\n"
                    "  rotate <0, -20, 0>\n"
                    "  translate <-1, 0, -3>\n"
                    "}\n"
                    "\n"
                    "text {\n"
                    "  ttf \"crystal.ttf\", \"Virtual World!\",\n"
                    "  1, // depth\n"
                    "  0  // spacing\n"
                    "  scale <1, 2, 1> // stretch it taller\n"
                    "  texture { Text_Tex }\n"
                    "  rotate <0, -30, 0>\n"
                    "  translate <-3, 0, 3>\n"
                    "}\n"
                    "\n"
                    "\n"
                    "plane { y, 0 pigment { color rgb <0.7,0.5,0.3> } }\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-10---Scene-File-Header ()
  (interactive
    (insert (concat
                    "// Persistence of Vision Ray Tracer Scene Description File\n"
                    "// File: .pov\n"
                    "// Vers: 3.5\n"
                    "// Desc:\n"
                    "// Date:\n"
                    "// Auth:\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-20---Standard-includes ()
  (interactive
    (insert (concat
                    "// ==== Standard POV-Ray Includes ====\n"
                    "#include \"colors.inc\"	  // Standard Color definitions\n"
                    "#include \"textures.inc\"	  // Standard Texture definitions\n"
                    "#include \"functions.inc\"  // internal functions usable in user defined functions\n"
             )
    )
  )
)

(defun pov-im-30---Additional-includes ()
  (interactive
    (insert (concat
                    "// ==== Additional Includes ====\n"
                    "// Don't have all of the following included at once, it'll cost memory and time\n"
                    "// to parse!\n"
                    "// --- general include files ---\n"
                    "#include \"arrays.inc\"     // macros for manipulating arrays\n"
                    "#include \"chars.inc\"      // A complete library of character objects, by Ken Maeno\n"
                    "#include \"consts.inc\"     // Various constants and alias definitions\n"
                    "#include \"debug.inc\"      // contains various macros for debugging scene files\n"
                    "#include \"logo.inc\"       // The official POV-Ray Logo in various forms\n"
                    "#include \"math.inc\"       // general math functions and macros\n"
                    "#include \"rad_def.inc\"    // Some common radiosity settings\n"
                    "#include \"rand.inc\"       // macros for generating random numbers\n"
                    "#include \"shapes.inc\"     // macros for generating various shapes\n"
                    "#include \"shapes2.inc\"    // some not built in basic shapes\n"
                    "#include \"shapesq.inc\"    // Pre-defined quartic shapes\n"
                    "#include \"skies.inc\"      // Ready defined sky spheres\n"
                    "#include \"stars.inc\"      // Some star fields\n"
                    "#include \"strings.inc\"    // macros for generating and manipulating text strings\n"
                    "#include \"sunpos.inc\"     // macro for sun position on a given date, time, and location on earth\n"
                    "#include \"transforms.inc\" // transformation macros \n"
                    "\n"
                    "// --- textures ---\n"
                    "#include \"finish.inc\"     // Some basic finishes\n"
                    "#include \"glass.inc\"      // Glass textures/interiors\n"
                    "#include \"golds.inc\"      // Gold textures\n"
                    "#include \"metals.inc\"     // Metallic pigments, finishes, and textures\n"
                    "#include \"stones.inc\"     // Binding include-file for STONES1 and STONES2\n"
                    "#include \"stones1.inc\"    // Great stone-textures created by Mike Miller\n"
                    "#include \"stones2.inc\"    // More, done by Dan Farmer and Paul Novak\n"
                    "#include \"woodmaps.inc\"   // Basic wooden colormaps\n"
                    "#include \"woods.inc\"      // Great wooden textures created by Dan Farmer and Paul Novak\n"
             )
    )
  )
)

(defun pov-im--arrays.inc- ()
  (interactive
    (insert (concat
                    "// array manipulation macros\n"
                    "// \n"
                    "#include \"arrays.inc\"\n"
             )
    )
  )
)

(defun pov-im--colors.inc- ()
  (interactive
    (insert (concat
                    "// Standard pre-defined colors\n"
                    "//\n"
                    "#include \"colors.inc\"\n"
             )
    )
  )
)

(defun pov-im--consts.inc- ()
  (interactive
    (insert (concat
                    "// various numerical constants\n"
                    "// \n"
                    "#include \"consts.inc\"\n"
             )
    )
  )
)

(defun pov-im--finish.inc- ()
  (interactive
    (insert (concat
                    "// some standard finishes\n"
                    "#include \"finish.inc\"\n"
                    "\n"
                    "  finish {\n"
                    "    Dull               // large, soft highlight\n"
                    "    //Shiny            // small, tight highlight\n"
                    "\n"
                    "    //Phong_Dull       // less \"realistic\" but sometimes useful phong highlight\n"
                    "    //Phong_Shiny      // same in \"shiny\"\n"
                    "\n"
                    "    //Glossy           // very tight highlights and some reflection\n"
                    "    //Phong_Glossy     // phong version of the same\n"
                    "\n"
                    "    //Luminous         // totally ambient not influenced by light sources (for example sky)\n"
                    "    //Mirror           // perfect mirror without highlights\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im--functions.inc- ()
  (interactive
    (insert (concat
                    "// internal functions usable in user defined functions\n"
                    "// for example for isosurfaces and function pattern\n"
                    "#include \"functions.inc\"\n"
             )
    )
  )
)

(defun pov-im--glass.inc- ()
  (interactive
    (insert (concat
                    "// various glass finishes, colors and interiors\n"
                    "//\n"
                    "#include \"glass.inc\"\n"
             )
    )
  )
)

(defun pov-im--math.inc- ()
  (interactive
    (insert (concat
                    "// general math functions and macros\n"
                    "// \n"
                    "#include \"math.inc\"\n"
             )
    )
  )
)

(defun pov-im--metals.inc- ()
  (interactive
    (insert (concat
                    "// several different gold colors, finishes and textures\n"
                    "#include \"golds.inc\"\n"
                    "\n"
                    "// various metal colors, finishes and textures\n"
                    "// brass, copper, chrome, silver\n"
                    "#include \"metals.inc\"\n"
             )
    )
  )
)

(defun pov-im--rand.inc- ()
  (interactive
    (insert (concat
                    "// random number generation macros\n"
                    "// \n"
                    "#include \"rand.inc\"\n"
             )
    )
  )
)

(defun pov-im--shapes.inc- ()
  (interactive
    (insert (concat
                    "// macros for generating various shapes\n"
                    "// \n"
                    "#include \"shapes.inc\"\n"
             )
    )
  )
)

(defun pov-im--shapes2.inc- ()
  (interactive
    (insert (concat
                    "// some not built in basic shapes\n"
                    "//\n"
                    "#include \"shapes2.inc\"\n"
                    "\n"
                    "object {\n"
                    "  Tetrahedron         // CSG-able tetrahedron shape\n"
                    "  //Octahedron        // CSG-able octahedron shape\n"
                    "  //Dodecahedron      // CSG-able dodecahedron shape\n"
                    "  //Icosahedron       // CSG-able icosahedron shape\n"
                    "  //Hexagon           // Hexagon extending along x-axis\n"
                    "  //Rhomboid          // 4-Sided Diamond\n"
                    "  //Pyramid           // non-CSG four-sided pyramid\n"
                    "  //Pyramid2          // CSG-able four-sided pyramid\n"
                    "  //Square_X          // square from <-1, -1> to <1, 1> in y-z-plane\n"
                    "  //Square_Y          // square from <-1, -1> to <1, 1> in x-z-plane\n"
                    "  //Square_Z          // square from <-1, -1> to <1, 1> in x-y-plane\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im--skies.inc- ()
  (interactive
    (insert (concat
                    "// some predefined skies\n"
                    "// \n"
                    "#include \"skies.inc\"\n"
             )
    )
  )
)

(defun pov-im--stones.inc- ()
  (interactive
    (insert (concat
                    "// a lot of stone textures\n"
                    "// T_Stone1 - T_Stone44\n"
                    "#include \"stones.inc\"\n"
                    "\n"
             )
    )
  )
)

(defun pov-im--strings.inc- ()
  (interactive
    (insert (concat
                    "// macros for generating and manipulating text strings\n"
                    "// \n"
                    "#include \"strings.inc\"\n"
             )
    )
  )
)

(defun pov-im--transforms.inc- ()
  (interactive
    (insert (concat
                    "// various geometric transformation macros\n"
                    "// \n"
                    "#include \"transforms.inc\"\n"
             )
    )
  )
)

(defun pov-im--woods.inc- ()
  (interactive
    (insert (concat
                    "// various (mostly layered) wood textures\n"
                    "// T_Wood1 - T_Wood35\n"
                    "#include \"woods.inc\"\n"
             )
    )
  )
)

(defun pov-im-10---background ()
  (interactive
    (insert (concat
                    "// set a color of the background (sky)\n"
                    "background { color rgb <0.1, 0.3, 0.8> }\n"
             )
    )
  )
)

(defun pov-im-20---global_settings ()
  (interactive
    (insert (concat
                    "// place all settings of globally influenced features here\n"
                    "global_settings {\n"
                    "// [GLOBAL ITEM(S)]\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-30---adc_bailout ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets cutoff point of ray computation (1/255 default)\n"
                    "  adc_bailout 0.0039\n"
             )
    )
  )
)

(defun pov-im-40---ambient_light ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets an overall brightness/ambient light level in the scene\n"
                    "  ambient_light color rgb <1,1,1>\n"
             )
    )
  )
)

(defun pov-im-50---assumed_gamma ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets image gamma in relation to display gamma\n"
                    "  assumed_gamma 1.0\n"
             )
    )
  )
)

(defun pov-im-60---charset ()
  (interactive
    (insert (concat
                    "// used in global_settings, specify the assumed character set of all text strings\n"
                    "  charset ascii       // standard ascii (only 0 to 127 are valid)\n"
                    "  //charset utf8      // unicode\n"
                    "  //charset sys       // system specific charset\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-70---hf_gray_16 ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets image output to use grayscale only (TGA, PNG)\n"
                    "  hf_gray_16 // [on/off]\n"
             )
    )
  )
)

(defun pov-im-80---irid_wavelength ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets the color shifts for iridescence (irid {})\n"
                    "  irid_wavelength <0.25, 0.18, 0.14>\n"
             )
    )
  )
)

(defun pov-im-90---max_intersections ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets the maximum ray tracing intersection depth (1 or more) [64]\n"
                    " max_intersections 64\n"
             )
    )
  )
)

(defun pov-im-A0---max_trace_level ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets the maximum ray tracing bounce depth (1 or more) [5]\n"
                    "  max_trace_level 5\n"
             )
    )
  )
)

(defun pov-im-B0---noise_generator ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets the type of noise computation\n"
                    "// there are 3 available noise generators\n"
                    "// they can be used for single pigments, etc. too by adding\n"
                    "// noise_generator x to the definition\n"
                    "  //noise_generator 1 // old 3.1g plateaud noise\n"
                    "  //noise_generator 2 // fixed version of 3.1 noise\n"
                    "  noise_generator 3   // New 3.5 perlin noise\n"
             )
    )
  )
)

(defun pov-im-C0---number_of_waves ()
  (interactive
    (insert (concat
                    "// used in global_settings, sets how many instances of waves/ripples to use\n"
                    "  number_of_waves 10\n"
             )
    )
  )
)

(defun pov-im-Bounds-extents ()
  (interactive
    (insert (concat
                    "// Bounds extents functions\n"
                    "// give the extent of the object's bounding box\n"
                    "#declare Obj = sphere { 0, 1 }\n"
                    "#declare Min = min_extent (Obj);\n"
                    "#declare Max = max_extent (Obj);\n"
             )
    )
  )
)

(defun pov-im-Built-in-Ids ()
  (interactive
    (insert (concat
                    "// These identifiers are built in to POV-Ray, ready to use\n"
                    "//  #declare pi = 3.1415926535897932384626\n"
                    "//  #declare true = 1\n"
                    "//  #declare yes = 1\n"
                    "//  #declare on = 1\n"
                    "//  #declare false = 0\n"
                    "//  #declare no = 0\n"
                    "//  #declare off = 0\n"
                    "//  #declare u = <1,0>\n"
                    "//  #declare v = <0,1>\n"
                    "//  #declare x = <1,0,0>\n"
                    "//  #declare y = <0,1,0>\n"
                    "//  #declare z = <0,0,1>\n"
                    "//  #declare t = <0,0,0,1>\n"
                    "\n"
                    "// Other built in identifiers with no fixed values are\n"
                    "/*\n"
                    "clock           // current clock value in animations (0..1)\n"
                    "clock_delta     // clock step between frames\n"
                    "version         // version (default 3.5, can be changed with #version)\n"
                    "initial_clock   // start clock value (+KIn.n)\n"
                    "final_clock     // end clock value (+KFn.n)\n"
                    "initial_frame   // start frame number (+KFIn.n)\n"
                    "final_frame     // end frame number (+KFFn.n)\n"
                    "frame_number    // current frame number\n"
                    "clock_on        // 1(true) if animation\n"
                    "image_width     // width of current render\n"
                    "image_height    // height of current render\n"
                    "*/\n"
             )
    )
  )
)

(defun pov-im-Conditional-Expression ()
  (interactive
    (insert (concat
                    "// The conditional expression returns a value depending\n"
                    "// on whether an expression is true or false.\n"
                    "// In this case: if X is larger than Y, A is returned, otherwise B\n"
                    "\n"
                    "#declare Value=((X>Y)?A:B);\n"
                    "\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-Float-Functions ()
  (interactive
    (insert (concat
                    "// Various float functions\n"
                    "/*\n"
                    "// (---general---)\n"
                    "#declare X=div(A,B);            // A/B (integer)\n"
                    "#declare X=mod(A,B);            // A modulo B\n"
                    "#declare X=sqrt(A);             // square root of A\n"
                    "#declare X=exp(A);              // e^A\n"
                    "#declare X=log(A);              // Natural logarithm of A\n"
                    "#declare X=pow(A,B);            // A^B\n"
                    "#declare X=max(A,B,C, ...);     // Maximum of A, B, C, ...\n"
                    "#declare X=min(A,B,C, ...);     // Minimum of A, B, C, ...\n"
                    "#declare X=int(A);              // truncate A to integer part\n"
                    "#declare X=abs(A);              // Absolute value of A\n"
                    "#declare X=ceil(A);             // ceiling of A (smallest larger integer)\n"
                    "#declare X=floor(A);            // floor of A (largest smaller integer)\n"
                    "#declare X=defined(A);          // true if A is a defined identifier\n"
                    "// (---trigonometric---)\n"
                    "#declare X=cos(A);              // cosine of A\n"
                    "#declare X=cosh(A);             // hyperbolic cosine of A\n"
                    "#declare X=sin(A);              // sine of A\n"
                    "#declare X=sinh(A);             // hyperbolic sine of A\n"
                    "#declare X=tan(A);              // tangent of A\n"
                    "#declare X=tanh(A);             // hyperbolic tangent of A\n"
                    "#declare X=acos(A);             // arc-cosine of A\n"
                    "#declare X=acosh(A);            // hyperbolic Arc-cosine of A\n"
                    "#declare X=asin(A);             // arc-sine of A\n"
                    "#declare X=asinh(A);            // hyperbolic Arc-sines of A\n"
                    "#declare X=atan(A);             // arc-tangent of A\n"
                    "#declare X=atanh(A);            // hyperbolic Arc-tangent of A\n"
                    "#declare X=atan2(A,B);          // arc-tangent of (A/B)\n"
                    "#declare X=degrees(A);          // convert radians to degrees\n"
                    "#declare X=radians(A);          // convert degrees to radians\n"
                    "// (---arrays---)\n"
                    "#declare X=dimensions(A);       // Number of dimensions of Array A\n"
                    "#declare X=dimension_size(A,B); // Size of Array A in dimension B\n"
                    "*/\n"
             )
    )
  )
)

(defun pov-im-Logical-Operators ()
  (interactive
    (insert (concat
                    "// As logical operators there are '&' and '|' standing for\n"
                    "// logical AND and OR.\n"
                    "// They return arithmetic value 0 for false or 1 for true.\n"
                    "// Note that A and B are not handled bitwise but logical.\n"
                    "\n"
                    "#if (A & B)\n"
                    "  ... // do this only if both A and B are true\n"
                    "#end\n"
                    "\n"
                    "#if (A | B)\n"
                    "  ... // do this if either A or B is true\n"
                    "#end\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-Math-Operators ()
  (interactive
    (insert (concat
                    "//  ( ) ! * / + -\n"
                    "#declare MyResult = -2*B + (A*A - 4)\n"
             )
    )
  )
)

(defun pov-im-Relational-Operators ()
  (interactive
    (insert (concat
                    "// Relationals must be within parentheses\n"
                    "// Return arithmetic value 0 for false or 1 for true\n"
                    "//   <  <=  =  !=  >=  >\n"
                    "\n"
                    "#declare BallColor = pigment { red (clock > 0.5) } // black or red\n"
             )
    )
  )
)

(defun pov-im-Strings ()
  (interactive
    (insert (concat
                    "/* various string operations\n"
                    "asc(S1)          // Convert 1st character of S1 to ASCII value\n"
                    "chr(A)           // Convert extended ASCII value A to a 1 character string\n"
                    "concat(S1,S2)    // combine S1 and S2 into one long string\n"
                    "file_exists(S1)  // Search current and include directories for existence of file S1 (0 or 1)\n"
                    "str(A,L,P)       // Convert float A to string, at least L characters long,\n"
                    "                 // with P digits after the decimal point (if P is -1, make max)\n"
                    "strcmp(S1,S2)    // compare S1 to S2, return -1, 0, or +1 if S2 is <, =, > than S1\n"
                    "strlen(S1)       // Returns # of characters in string S1\n"
                    "strlwr(S1)       // Lower case of S1\n"
                    "substr(S1,P,L)   // Sub-string from S1, start at position P for length L\n"
                    "strupr(S1)       // Upper case of S1\n"
                    "val(S1)          // Convert string S1 to float\n"
                    "file_exists(S1)  // returns 1 if file S1 is found in library path\n"
                    "*/\n"
                    "/*\n"
                    "// some special control characters that can be used within strings\n"
                    "  \"\a\" Bell or alarm, 0x07\n"
                    "  \"\b\" Backspace, 0x08\n"
                    "  \"\f\" Form feed, 0x0C\n"
                    "  \"\n\" New line (line feed) 0x0A\n"
                    "  \"\r\" Carriage return 0x0D\n"
                    "  \"\t\" Horizontal tab 0x09\n"
                    "  \"\v\" Vertical tab 0x0B\n"
                    "  \"\0\" Null 0x00\n"
                    "  \"\\\" Backslash 0x5C\n"
                    "  \"\'\" Single quote 0x27\n"
                    "*/\n"
             )
    )
  )
)

(defun pov-im-Vector-Functions ()
  (interactive
    (insert (concat
                    "// vector functions\n"
                    "/*\n"
                    "#declare X=vaxis_rotate(A,B,F); // rotate A about B by F\n"
                    "#declare X=vcross(A,B);         // cross product of A and B\n"
                    "#declare X=vdot(A,B);           // dot product of A and B\n"
                    "#declare X=vrotate(A,B);        // rotate A about origin by B\n"
                    "#declare X=vnormalize(A);       // normalize vector A\n"
                    "#declare X=vturbulence(Lambda,Omega,Octaves, A);\n"
                    "                                // X+A gives a turbulated version of A\n"
                    "*/\n"
             )
    )
  )
)

(defun pov-im-Vector-components ()
  (interactive
    (insert (concat
                    "// extract each component of a vector:\n"
                    "\n"
                    "// for 3D vectors:\n"
                    "// #if (MyVector.x > 5) ...\n"
                    "// #if (MyVector.y > 5) ...\n"
                    "// #if (MyVector.z > 5) ...\n"
                    "\n"
                    "// in addition for 4D vectors:\n"
                    "// #if (MyVector.t > 5) ...\n"
                    "\n"
                    "// for 2D (UV) vectors:\n"
                    "// #if (MyVector.u > 5) ...\n"
                    "// #if (MyVector.v > 5) ...\n"
             )
    )
  )
)

(defun pov-im-inside ()
  (interactive
    (insert (concat
                    "// inside function tests if a certain point  \n"
                    "// is inside a specified object\n"
                    "\n"
                    "#declare Obj = sphere { 0, 1 }\n"
                    "#declare Point = <0.6, 0.5, 0.6>;\n"
                    "\n"
                    "// if point is inside the object:\n"
                    "#if ( inside(Obj, Point) )\n"
                    " //...\n"
                    "#end\n"
             )
    )
  )
)

(defun pov-im-rand ()
  (interactive
    (insert (concat
                    "// the rand() function creates a reproducible sequence \n"
                    "// of pseudo-random numbers between 0.0 and 1.0\n"
                    "\n"
                    "#declare R1 = seed(0);      // initialize random number streams\n"
                    "#declare R2 = seed(12345);\n"
                    "\n"
                    "// place sphere with random radius at random position\n"
                    "// use different random number streams for position and radius\n"
                    "sphere { <rand(R1), rand(R1), rand(R1)>, rand(R2) }\n"
             )
    )
  )
)

(defun pov-im-trace ()
  (interactive
    (insert (concat
                    "// trace function tests for intersection with a specified object\n"
                    "\n"
                    "#declare Obj = sphere { 0, 1 }\n"
                    "#declare Norm = <0, 0, 0>;\n"
                    "#declare Start = <0.5, 0.5, 1>;\n"
                    "#declare Pos = trace (\n"
                    "                  Obj,             // object to test\n"
                    "                  Start,           // starting point\n"
                    "                  -z,              // direction\n"
                    "                  Norm );          // normal\n"
                    "\n"
                    "// if intersection is found, normal differs from 0\n"
                    "#if (Norm.x != 0 | Norm.y != 0 | Norm.z != 0)\n"
                    " //...\n"
                    "#end\n"
             )
    )
  )
)

(defun pov-im-declare ()
  (interactive
    (insert (concat
                    "// note that #declare of float, vector and color require semicolon at the end\n"
                    "#declare MyColor = color rgb<0.7, 0.5, 0.3>;\n"
             )
    )
  )
)

(defun pov-im-default ()
  (interactive
    (insert (concat
                    "// sets the default texture that objects get when they have no texture specified\n"
                    "#default {\n"
                    "  texture { pigment {color red 1} finish{ambient 0.2} }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-if-example ()
  (interactive
    (insert (concat
                    "  #if (High_Quality)\n"
                    "    // This section is parsed if High_Quality is true\n"
                    "  #end // End of conditional part\n"
                    "// or you can use the else clause too\n"
                    "  #if (clock > 2)\n"
                    "    // This section is parsed if clock is > 2\n"
                    "  #else\n"
                    "    // This section is parsed if clock is <= 2\n"
                    "  #end // End of conditional part\n"
             )
    )
  )
)

(defun pov-im-if ()
  (interactive
    (insert (concat
                    "#if (expr)\n"
                    "  // true stuff happens here\n"
                    "#else\n"
                    "  // false stuff happens here\n"
                    "#end\n"
             )
    )
  )
)

(defun pov-im-ifdef-example ()
  (interactive
    (insert (concat
                    "  #ifdef (SphereFlake_Shape)\n"
                    "    // This section is  parsed if SphereFlake_Shape is declared\n"
                    "  #else\n"
                    "    // This section is  parsed if SphereFlake_Shape is NOT declared\n"
                    "    #declare SphereFlake_Shape = sphere {0,1} // make a default shape\n"
                    "  #end // End of conditional part\n"
             )
    )
  )
)

(defun pov-im-ifdef ()
  (interactive
    (insert (concat
                    "#ifdef (ident)\n"
                    "  // true stuff happens here\n"
                    "#else\n"
                    "  // false stuff happens here\n"
                    "#end\n"
             )
    )
  )
)

(defun pov-im-local ()
  (interactive
    (insert (concat
                    "// similar to #declare, but limited to the current scene file/macro\n"
                    "// note that #local of float, vector and color require semicolon at the end\n"
                    "#local MyColor = color rgb<0.7, 0.5, 0.3>;\n"
             )
    )
  )
)

(defun pov-im-macro ()
  (interactive
    (insert (concat
                    "// macros can have parameters and are useful for 'programming' scenes\n"
                    "#macro Spheres(A, B, C)\n"
                    "  sphere { A, C }\n"
                    "  sphere { B, C }\n"
                    "#end\n"
                    "\n"
                    "// this macro can be called for example with\n"
                    "union { Spheres(x, 2*x, 0.5) }\n"
             )
    )
  )
)

(defun pov-im-messages ()
  (interactive
    (insert (concat
                    "// display user defined messages in the Povray output\n"
                    "  #debug      \"During scene parsing, general info message\"\n"
                    "  #warning    \"During scene parsing\"\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-switch-example ()
  (interactive
    (insert (concat
                    "// Let's make some constant names\n"
                    "#declare CS_Medium = 1;\n"
                    "#declare CS_Hard = 3;\n"
                    "#declare CS_Easy = 5;\n"
                    "\n"
                    "// Let the user choose the method to use\n"
                    "#declare Complexity_Switch = CS_Medium; // or CS_Easy or CS_Hard\n"
                    "\n"
                    "// Do something dependent on the user's choice\n"
                    "#switch (Complexity_Switch)\n"
                    "\n"
                    "  #case (CS_Easy)\n"
                    "    // This statement is done if Complexity_Switch is CS_Easy\n"
                    "    #declare MyShape = box { -<1, 1, 1>, <1, 1, 1> }\n"
                    "  #break // End of this case section\n"
                    "\n"
                    "  #range (CS_Medium, CS_Hard)\n"
                    "    // This statement is done if Complexity_Switch is CS_Medium\n"
                    "    // or CS_Hard or anything in between\n"
                    "    #declare MyShape = torus { 1, 0.5 }\n"
                    "  #break // End of this range section\n"
                    "\n"
                    "  #else\n"
                    "    // This statement is done if none of the above match\n"
                    "    #declare MyShape = sphere { <0, 0, 0>, 1 }\n"
                    "\n"
                    "#end // End of switch statement\n"
             )
    )
  )
)

(defun pov-im-switch ()
  (interactive
    (insert (concat
                    "#switch (expr)\n"
                    "  #case (expr)\n"
                    "  #break\n"
                    "  #range (expr,expr)\n"
                    "  #break\n"
                    "  #else\n"
                    "#end\n"
             )
    )
  )
)

(defun pov-im-undef ()
  (interactive
    (insert (concat
                    "// identifiers can be destroyed with\n"
                    "#undef MyColor\n"
             )
    )
  )
)

(defun pov-im-version ()
  (interactive
    (insert (concat
                    "#declare Temp_Vers = version    // Save original value of #version\n"
                    "#version 3.0;                   // Change to 3.0 mode\n"
                    "// Version 3.0 stuff goes here ...\n"
                    "#version Temp_Vers              // Restore original #version value\n"
             )
    )
  )
)

(defun pov-im-while ()
  (interactive
    (insert (concat
                    "// Create 10 balls along X axis, from 0 to 9\n"
                    "#declare BallCount = 0;\n"
                    "#while (BallCount < 10)\n"
                    "  sphere {\n"
                    "    <BallCount,0,0>,  // NOTE: <0,0,0>, <1,0,0>, <2,0,0>, etc.\n"
                    "    0.5\n"
                    "  }\n"
                    "  #declare BallCount = BallCount+1;   // increment our counter\n"
                    "#end\n"
             )
    )
  )
)

(defun pov-im-10---clock ()
  (interactive
    (insert (concat
                    "  // spin the torus around the X axis, as \"clock\" goes from 0 to 1\n"
                    "  // the torus will rotate one complete 360 degree revolution.\n"
                    "  // Note that the texture is put on BEFORE we rotate so it rotates too.\n"
                    "  torus { 1,0.2 pigment {color red 1} rotate clock*x*360}\n"
             )
    )
  )
)

(defun pov-im-20---additional-animation-variables ()
  (interactive
    (insert (concat
                    "// additional animation variables\n"
                    "#declare X=clock_delta;       // clock step between frames\n"
                    "#declare X=initial_clock;     // clock value of first rendered frame (+KIn.n)\n"
                    "#declare X=final_clock;       // clock value of last rendered frame (+KFn.n)\n"
                    "#declare X=initial_frame;     // initial frame number (+KFIn).\n"
                    "#declare X=final_frame;       // final frame number (+KFIn).\n"
                    "#declare X=clock_on;          // true if scene is rendered as an animation\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-10---rgb ()
  (interactive
    (insert (concat
                    "// rgb - Red Green Blue color shortcut\n"
                    "color rgb <0.2, 0.3, 0.4>\n"
             )
    )
  )
)

(defun pov-im-20---rgbf ()
  (interactive
    (insert (concat
                    "// rgbf - Red Green Blue Filter color shortcut\n"
                    "color rgbf <0.2, 0.3, 0.4, 0.9>\n"
             )
    )
  )
)

(defun pov-im-30---rgbt ()
  (interactive
    (insert (concat
                    "// rgbt - Red Green Blue Transmit color shortcut\n"
                    "color rgbt <0.2, 0.3, 0.4, 0.9>\n"
             )
    )
  )
)

(defun pov-im-40---rgbft ()
  (interactive
    (insert (concat
                    "// rgbft - Red Green Blue Filter Transmit color shortcut\n"
                    "// NOTE: filter + transmit should always equal 1.0\n"
                    "color rgbft <0.5, 0.5, 0.5, 0.4, 0.6>\n"
             )
    )
  )
)

(defun pov-im-50---red ()
  (interactive
    (insert (concat
                    "color red 1\n"
             )
    )
  )
)

(defun pov-im-60---green ()
  (interactive
    (insert (concat
                    "color green 1\n"
             )
    )
  )
)

(defun pov-im-70---blue ()
  (interactive
    (insert (concat
                    "color blue 1\n"
             )
    )
  )
)

(defun pov-im-80---filter ()
  (interactive
    (insert (concat
                    "color red 1 filter 1\n"
             )
    )
  )
)

(defun pov-im-90---transmit ()
  (interactive
    (insert (concat
                    "color green 1 transmit 1\n"
             )
    )
  )
)

(defun pov-im-A0---color-components ()
  (interactive
    (insert (concat
                    "// extract each component of a color:\n"
                    "// #if (MyColor.red < 0.5) ...\n"
                    "// #if (MyColor.green < 0.5) ...\n"
                    "// #if (MyColor.blue < 0.5) ...\n"
                    "// #if (MyColor.filter < 0.5) ...\n"
                    "// #if (MyColor.transmit < 0.5) ...\n"
             )
    )
  )
)

(defun pov-im-10---Fog-definition ()
  (interactive
    (insert (concat
                    "// set global atmospheric fog effect in the scene.\n"
                    "// at the fog distance, there will be 63% visibility\n"
                    "fog {\n"
                    "  fog_type 1               // 1=constant, 2=ground_fog\n"
                    "  distance 10\n"
                    "  color Gray               // can also have 'filter' and 'transmit'\n"
                    "  // (---turbulence---)\n"
                    "  //turbulence <0.5, 0.5, 1.0>\n"
                    "  //turb_depth 0.5\n"
                    "  //omega 0.5\n"
                    "  //lambda 2.0\n"
                    "  //octaves 6\n"
                    "  // (---ground fog---)\n"
                    "  //fog_offset 0.5         // height of constant fog\n"
                    "  //fog_alt 0.5            // at fog_offset+fog_alt: density=25%\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-20---Ambient-fog ()
  (interactive
    (insert (concat
                    "fog {\n"
                    "  fog_type 1\n"
                    "  distance 100\n"
                    "  color SkyBlue\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-30---Ground-fog ()
  (interactive
    (insert (concat
                    "fog {\n"
                    "  fog_type   2\n"
                    "  distance   10\n"
                    "  color      rgb 0.6 // gray\n"
                    "  fog_offset 0.1\n"
                    "  fog_alt    0.2\n"
                    "  turbulence 0.8\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-40---Atmospheric-media ()
  (interactive
    (insert (concat
                    "// atmospheric media can be generated by adding a media statement\n"
                    "// to the scene not attached to any specific object\n"
                    "\n"
                    "media {    // atmospheric media sample\n"
                    "  intervals 10\n"
                    "  scattering { 1, rgb 0.03 }\n"
                    "  samples 1, 10\n"
                    "  confidence 0.9999\n"
                    "  variance 1/1000\n"
                    "  ratio 0.9\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-50---sky_sphere ()
  (interactive
    (insert (concat
                    "// Create an infinite sphere around scene and allow any pigment on it\n"
                    "sky_sphere {\n"
                    "  pigment {\n"
                    "    gradient y\n"
                    "    color_map { [0.0 color rgb <0.7,0.7,1.0>] [1.0 color blue 0.5] }\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-60---rainbow ()
  (interactive
    (insert (concat
                    "// create a rainbow arc\n"
                    "rainbow {                      \n"
                    "  angle 30         // degrees\n"
                    "  width 6          // degrees\n"
                    "  distance 1000\n"
                    "  direction <0,-0.5,1>    \n"
                    "  jitter 0.01\n"
                    "  arc_angle 180    // degrees\n"
                    "  falloff_angle 60 // degrees\n"
                    "  // up <UP>\n"
                    "  color_map {\n"
                    "    [0.000 color rgbf <1.0, 0.5, 1.0, 1.0> ]\n"
                    "    [0.100 color rgbf <1.0, 0.5, 1.0, 0.8> ]\n"
                    "    [0.214 color rgbf <0.5, 0.5, 1.0, 0.8> ]\n"
                    "    [0.328 color rgbf <0.2, 0.2, 1.0, 0.8> ]\n"
                    "    [0.442 color rgbf <0.2, 1.0, 1.0, 0.8> ]\n"
                    "    [0.556 color rgbf <0.2, 1.0, 0.2, 0.8> ]\n"
                    "    [0.670 color rgbf <1.0, 1.0, 0.2, 0.8> ]\n"
                    "    [0.784 color rgbf <1.0, 0.5, 0.2, 0.8> ]\n"
                    "    [0.900 color rgbf <1.0, 0.2, 0.2, 0.8> ]\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-70---Global-photons ()
  (interactive
    (insert (concat
                    "// adding a photon{} block to global_settings activates photon mapping.\n"
                    "// photons also need to be adjusted for light sources and objects.\n"
                    "global_settings {\n"
                    "  photons {\n"
                    "    spacing 0.01                 // specify the density of photons\n"
                    "    //count 100000               // alternatively use a total number of photons\n"
                    "\n"
                    "    //gather min, max            // amount of photons gathered during render [20, 100]\n"
                    "    //media max_steps [,factor]  // media photons\n"
                    "    //jitter 1.0                 // jitter phor photon rays\n"
                    "    //max_trace_level 5          // optional separate max_trace_level\n"
                    "    //adc_bailout 1/255          // see global adc_bailout\n"
                    "    //save_file \"filename\"       // save photons to file\n"
                    "    //load_file \"filename\"       // load photons from file\n"
                    "    //autostop 0                 // photon autostop option\n"
                    "    //radius 10                  // manually specified search radius\n"
                    "    // (---Adaptive Search Radius---)\n"
                    "    //steps 1\n"
                    "    //expand_thresholds 0.2, 40\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-80---Light-source-photons ()
  (interactive
    (insert (concat
                    "// photon block for a light source\n"
                    "  photons {\n"
                    "    refraction on\n"
                    "    reflection on\n"
                    "    //area_light\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-90---Object-photons ()
  (interactive
    (insert (concat
                    "// photon block for an object\n"
                    "  photons{\n"
                    "    target 1.0          // spacing multiplier for photons hitting the object\n"
                    "    refraction on\n"
                    "    reflection on\n"
                    "    //collect off       // ignore photons\n"
                    "    //pass_through      // do not influence photons\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-A0---radiosity ()
  (interactive
    (insert (concat
                    "// radiosity (global illumination) settings\n"
                    "global_settings {\n"
                    "  radiosity {\n"
                    "    pretrace_start 0.08           // start pretrace at this size\n"
                    "    pretrace_end   0.04           // end pretrace at this size\n"
                    "    count 35                      // higher -> higher quality (1..1600) [35]\n"
                    "    nearest_count 5               // higher -> higher quality (1..10) [5]\n"
                    "    error_bound 1.8               // higher -> smoother, less accurate [1.8]\n"
                    "    recursion_limit 3             // how much interreflections are calculated (1..5+) [3]\n"
                    "    low_error_factor .5           // reduce error_bound during last pretrace step\n"
                    "    gray_threshold 0.0            // increase for weakening colors (0..1) [0]\n"
                    "    minimum_reuse 0.015           // reuse of old radiosity samples [0.015]\n"
                    "    brightness 1                  // brightness of radiosity effects (0..1) [1]\n"
                    "\n"
                    "    adc_bailout 0.01/2\n"
                    "    //normal on                   // take surface normals into account [off]\n"
                    "    //media on                    // take media into account [off]\n"
                    "    //save_file \"file_name\"       // save radiosity data\n"
                    "    //load_file \"file_name\"       // load saved radiosity data\n"
                    "    //always_sample off           // turn sampling in final trace off [on]\n"
                    "    //max_sample 1.0              // maximum brightness of samples\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-A-typical-camera ()
  (interactive
    (insert (concat
                    "// perspective (default) camera\n"
                    "camera {\n"
                    "  location  <0.0, 2.0, -5.0>\n"
                    "  look_at   <0.0, 0.0,  0.0>\n"
                    "  right     x*image_width/image_height\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-Camera-definition ()
  (interactive
    (insert (concat
                    "// set viewer's position in the scene\n"
                    "camera {\n"
                    "// (---camera types---)\n"
                    "//  perspective (default) | orthographic | fisheye |\n"
                    "//  ultra_wide_angle | omnimax | panoramic | cylinder 1 | spherical\n"
                    "\n"
                    "  location  <0.0, 1.0, -6.0>   // position of camera <X,Y,Z>\n"
                    "  direction 2.0*z              // which way are we looking <X,Y,Z> & zoom\n"
                    "//  sky       y                // for tilting the camera\n"
                    "  up        y                  // which way is +up <X,Y,Z> (aspect ratio with x)\n"
                    "  right x*image_width/image_height\n"
                    "                               // which way is +right <X,Y,Z> (aspect ratio with y)\n"
                    "  look_at   <0.0, 0.0,  0.0>   // point center of view at this point <X,Y,Z>\n"
                    "\n"
                    "//  angle 67                   // overrides \"direction\" with specific angle\n"
                    "//  normal { ripples 0.2 }     // perturb the camera lens with a pattern\n"
                    "\n"
                    "// (---focal blur extras---)\n"
                    "// aperture 0.2                // [0...N] larger is narrower depth of field (blurrier)\n"
                    "// blur_samples 4              // number of rays per pixel for sampling\n"
                    "// focal_point <0,0,0>         // point that is in focus <X,Y,Z>\n"
                    "// confidence 0.9              // [0...1] when to move on while sampling (smaller is less accurate)\n"
                    "// variance 1/128              // [0...1] how precise to calculate (smaller is more accurate)\n"
                    "\n"
                    "// (---spherical camera---)\n"
                    "// angle 360                   // horizontal field of view (in degrees and...)\n"
                    "//       180                   // vertical field of view (...are aspect ratios)\n"
                    "}\n"
                    "\n"
                    "\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-cylinder-camera ()
  (interactive
    (insert (concat
                    "// cylinder lens for axis distorted field of view\n"
                    "// 1 is vertical cylinder, fixed viewpoint\n"
                    "// 2 is horizontal cylinder, fixed viewpoint\n"
                    "// 3 is vertical cylinder, viewpoint moves along the cylinder's axis\n"
                    "// 4 is horizontal cylinder, viewpoint moves along the cylinder's axis\n"
                    "camera {\n"
                    "  cylinder 1            // types 1, 2 ,3, 4\n"
                    "  location <0,0,-1>     // position\n"
                    "  right x*5             // horizontal aspect or field of view\n"
                    "  up y*2                // vertical aspect or field of view\n"
                    "  look_at  <0,0,0>      // view\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-fisheye-camera ()
  (interactive
    (insert (concat
                    "// fisheye lens for wide round field of view (up to 360 degrees)\n"
                    "camera {\n"
                    "  fisheye\n"
                    "  location <0,0,-1>     // position\n"
                    "  look_at  <0,0,0>      // view\n"
                    "  angle 180             // field\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-focal-blur-camera ()
  (interactive
    (insert (concat
                    "// focal blur camera\n"
                    "camera {\n"
                    "  location  <0.0, 2.0, -5.0>\n"
                    "  look_at   <0.0, 0.0,  0.0>\n"
                    "  right     x*image_width/image_height\n"
                    "  aperture 1.0           // [0...N] larger is narrower depth of field (blurrier)\n"
                    "  blur_samples 10        // number of rays per pixel for sampling\n"
                    "  focal_point <0,1,2>    // point that is in focus <X,Y,Z>\n"
                    "  confidence 0.95           // [0...<1] when to move on while sampling (smaller is less accurate)\n"
                    "  variance 1/200            // [0...1] how precise to calculate (smaller is more accurate)\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-normal-perturbed-camera ()
  (interactive
    (insert (concat
                    "// normal perturbed camera\n"
                    "camera {\n"
                    "  location  <0.0, 2.0, -5.0>\n"
                    "  look_at   <0.0, 0.0,  0.0>\n"
                    "  right     x*image_width/image_height\n"
                    "  normal {\n"
                    "   wrinkles 0.2 scale 0.3\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-omnimax-camera ()
  (interactive
    (insert (concat
                    "// omnimax lens for 180 degree equirectangular field of view (angle unused)\n"
                    "camera {\n"
                    "  omnimax\n"
                    "  location <0,0,-1>                    // position\n"
                    "  look_at  <0,0,0>                     // view\n"
                    "  right    x*image_width/image_height  // aspect\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-orthographic-camera ()
  (interactive
    (insert (concat
                    "// orthographic projection using parallel camera rays\n"
                    "// Could be used to render a planar image map, for example\n"
                    "camera {\n"
                    "  orthographic\n"
                    "  location <0,0,-1>    // position & direction of view\n"
                    "  look_at  <0,0,0>\n"
                    "  right 1*x            // horizontal size of view\n"
                    "  up 1*y               // vertical size of view\n"
                    "}\n"
                    "/*\n"
                    "box {                  // this box fits exactly in view\n"
                    "  <-0.5, -0.5, 0>, <0.5, 0.5, 0>\n"
                    "  texture { pigment { rgb <1, 0, 1> } }\n"
                    "}*/\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-panoramic-camera ()
  (interactive
    (insert (concat
                    "// panoramic lens for wide field of view with less distortion\n"
                    "camera {\n"
                    "  panoramic\n"
                    "  location <0, 0, -1>                  // position\n"
                    "  look_at  <0, 0,  0>                  // view\n"
                    "  right    x*image_width/image_height  // aspect\n"
                    "  angle    120                         // field (greater than 180 degrees possible)\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-perspective-camera ()
  (interactive
    (insert (concat
                    "// perspective (default, not required) camera\n"
                    "camera {\n"
                    "  perspective\n"
                    "  location  <0, 0, -1>\n"
                    "  look_at   <0, 0,  0>\n"
                    "  right     x*image_width/image_height  // aspect\n"
                    "  // direction z                        // direction and zoom\n"
                    "  // angle 67                           // field (overides direction zoom)\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-sky-camera ()
  (interactive
    (insert (concat
                    "// sky (tilt) camera\n"
                    "camera {\n"
                    "  location  <0.0, 2.0, -5.0>\n"
                    "  sky       <1.0, 0.5,  0.0>        // direction of the camera \"sky\"\n"
                    "  look_at   <0.0, 0.0,  0.0>\n"
                    "  right     x*image_width/image_height\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-spherical-camera ()
  (interactive
    (insert (concat
                    "// spherical lens for sphere field of view (mappable to a sphere)\n"
                    "camera {\n"
                    "  spherical\n"
                    "  location <0,0,0>      // position\n"
                    "  look_at  <0,0,1>      // view\n"
                    "  angle 360             // horizontal degrees\n"
                    "        180             // vertical degrees\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-ultra_wide_angle-camera ()
  (interactive
    (insert (concat
                    "// ultra_wide_angle lens for wide, rectangular field of view\n"
                    "camera {\n"
                    "  ultra_wide_angle\n"
                    "  location <0, 1, -5>               // position\n"
                    "  look_at  <0, 0, 0>                // view\n"
                    "  right x*image_width/image_height  // aspect\n"
                    "  angle 90                          // field\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-Light-definition ()
  (interactive
    (insert (concat
                    "// general light definition\n"
                    "light_source {\n"
                    "  <10, 10, 10>      // position of the light source\n"
                    "  color rgb 1.0     // color of the light\n"
                    "  // spotlight\n"
                    "  // cylinder\n"
                    "  // parallel\n"
                    "  // area_light <AXIS1>, <AXIS2>, SIZE1, SIZE2\n"
                    "  // (---for spotlight/cylinder---)\n"
                    "  // radius FLOAT\n"
                    "  // falloff FLOAT\n"
                    "  // tightness FLOAT\n"
                    "  // point_at <VECTOR>   // for spotlight/cylinder/parallel\n"
                    "  // (---for area_light---)\n"
                    "  // adaptive FLOAT\n"
                    "  // jitter FLOAT\n"
                    "  // circular\n"
                    "  // orient\n"
                    "  // (---other modifiers---)\n"
                    "  // looks_like { OBJECT }\n"
                    "  // fade_distance FLOAT\n"
                    "  // fade_power FLOAT\n"
                    "  // media_attenuation BOOL\n"
                    "  // media_interaction BOOL\n"
                    "  // shadowless\n"
                    "}\n"
                    "\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-area_light ()
  (interactive
    (insert (concat
                    "// An area light (creates soft shadows)\n"
                    "// WARNING: This special light can significantly slow down rendering times!\n"
                    "light_source {\n"
                    "  0*x                 // light's position (translated below)\n"
                    "  color rgb 1.0       // light's color\n"
                    "  area_light\n"
                    "  <8, 0, 0> <0, 0, 8> // lights spread out across this distance (x * z)\n"
                    "  4, 4                // total number of lights in grid (4x*4z = 16 lights)\n"
                    "  adaptive 0          // 0,1,2,3...\n"
                    "  jitter              // adds random softening of light\n"
                    "  circular            // make the shape of the light circular\n"
                    "  orient              // orient light\n"
                    "  translate <40, 80, -40>   // <x y z> position of light\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-cylindrical-light ()
  (interactive
    (insert (concat
                    "// create a point \"spotlight\" (cylindrical directed) light source\n"
                    "light_source {\n"
                    "  0*x                     // light's position (translated below)\n"
                    "  color rgb <1,1,1>       // light's color\n"
                    "  spotlight               // this kind of light source\n"
                    "  cylinder                // this variation\n"
                    "  translate <40, 80, -40> // <x y z> position of light\n"
                    "  point_at <0, 0, 0>      // direction of spotlight\n"
                    "  radius 5                // hotspot (inner, in degrees)\n"
                    "  tightness 50            // tightness of falloff (1...100) lower is softer, higher is tighter\n"
                    "  falloff 8               // intensity falloff radius (outer, in degrees)\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-light-fading ()
  (interactive
    (insert (concat
                    "// light_source { ...\n"
                    "  // put this inside a light_source to add light fading\n"
                    "  fade_distance 10\n"
                    "  fade_power 2\n"
             )
    )
  )
)

(defun pov-im-light_group ()
  (interactive
    (insert (concat
                    "// a light group makes certain light sources only\n"
                    "// influencing certain objects\n"
                    "light_group {\n"
                    "  light_source {          // light sourc(es) of this group\n"
                    "    <100, 100, 100>\n"
                    "    color rgb y\n"
                    "  }\n"
                    "  sphere {                // objects illuminated by those light sources\n"
                    "    <0, -1, 0>,0.5\n"
                    "    pigment { color White }\n"
                    "  }\n"
                    "  //global_lights         // add this to make all global lights\n"
                    "                          // also illuminating this light group\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-looks_like ()
  (interactive
    (insert (concat
                    "// light_source { ...\n"
                    "  // put this inside a light_source to give it a visible appearance\n"
                    "  looks_like { sphere { 0*x, 5 pigment { Yellow } } }\n"
             )
    )
  )
)

(defun pov-im-parallel ()
  (interactive
    (insert (concat
                    "// light_source { ...\n"
                    "  // put this inside a light_source to make it parallel\n"
                    "  parallel\n"
                    "  point_at <1, 0, 0>\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-point-light ()
  (interactive
    (insert (concat
                    "// create a regular point light source\n"
                    "light_source {\n"
                    "  0*x                  // light's position (translated below)\n"
                    "  color rgb <1,1,1>    // light's color\n"
                    "  translate <-20, 40, -20>\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-projected_through ()
  (interactive
    (insert (concat
                    "// light_source { ...\n"
                    "  // put this inside a light_source to give the light beam\n"
                    "  // the shape of the object\n"
                    "  projected_through { object { ... } }\n"
             )
    )
  )
)

(defun pov-im-spotlight ()
  (interactive
    (insert (concat
                    "// create a point \"spotlight\" (conical directed) light source\n"
                    "light_source {\n"
                    "  0*x                     // light's position (translated below)\n"
                    "  color rgb <1,1,1>       // light's color\n"
                    "  spotlight               // this kind of light source\n"
                    "  translate <40, 80, -40> // <x y z> position of light\n"
                    "  point_at <0, 0, 0>      // direction of spotlight\n"
                    "  radius 5                // hotspot (inner, in degrees)\n"
                    "  tightness 50            // tightness of falloff (1...100) lower is softer, higher is tighter\n"
                    "  falloff 8               // intensity falloff radius (outer, in degrees)\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-bicubic_patch ()
  (interactive
    (insert (concat
                    "// 3D curved FINITE (no CSG) surface created from a mesh of triangles\n"
                    "bicubic_patch {\n"
                    "  type 1 // patch_type (0..1)\n"
                    "  // 0 = Bezier patch, just store the triangular vertices\n"
                    "  // 1 = Bezier patch, store all plane equations defined by\n"
                    "  //       the triangulation of the patch into sub patches\n"
                    "  //       (faster, uses more memory)\n"
                    "   flatness 0.1 // flatness value\n"
                    "  //       flatness_value = 0.0 to 1.0, with higher values\n"
                    "  //        giving flatter, less smooth results\n"
                    "  u_steps 3 // # of triangles to subdivide (1-5)\n"
                    "  v_steps 3 // # of triangles to subdivide (1-5)\n"
                    "  <0, 0, 2> <1, 0, 0> <2, 0, 0> <3, 0, -2>\n"
                    "  <0, 1, 0> <1, 1, 0> <2, 1, 0> <3, 1,  0>\n"
                    "  <0, 2, 0> <1, 2, 0> <2, 2, 0> <3, 2,  0>\n"
                    "  <0, 3, 2> <1, 3, 0> <2, 3, 0> <3, 3, -2>\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-blob ()
  (interactive
    (insert (concat
                    "// create a smooth blobby shape\n"
                    "#declare RadiusVal   = 1.0; // (0 < RadiusVal) outer sphere of influence on other components\n"
                    "#declare StrengthVal = 1.0; // (+ or -) strength of component's radiating density\n"
                    "\n"
                    "blob {\n"
                    "  // threshold (0.0 < threshold <= StrengthVal) surface falloff threshold #\n"
                    "  threshold 0.6\n"
                    "  sphere { < 0.75,   0,    0>, RadiusVal, StrengthVal }\n"
                    "  sphere { <-0.375,  0.65, 0>, RadiusVal, StrengthVal }\n"
                    "  sphere { <-0.375, -0.65, 0>, RadiusVal, StrengthVal }\n"
                    "  cylinder { -z, +z, RadiusVal, StrengthVal }\n"
                    "  // [sturm]\n"
                    "  scale 2\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-box ()
  (interactive
    (insert (concat
                    "// create a box that extends between the 2 specified points\n"
                    "box {\n"
                    "  <-1, -1, -1>  // one corner position <X1 Y1 Z1>\n"
                    "  < 1,  1,  1>  // other corner position <X2 Y2 Z2>\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-cone ()
  (interactive
    (insert (concat
                    "// clipped conical shape\n"
                    "// cone { <END1>, RADIUS1, <END2>, RADIUS2 [open] }\n"
                    "// Where <END1> and <END2> are vectors defining the x,y,z\n"
                    "// coordinates of the center of each end of the cone\n"
                    "// and RADIUS1 and RADIUS2 are float values for the radii\n"
                    "// of those ends.  open, if present, cone is hollow, else capped\n"
                    "cone {\n"
                    "  1*y,  0.0,\n"
                    "  -1*y, 1.0\n"
                    "  // open\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-cubic ()
  (interactive
    (insert (concat
                    "// create a 3rd order infinite polynomial surface\n"
                    "cubic {\n"
                    "  <\n"
                    "// x^3,      x^2y,     x^2z,     x^2,\n"
                    "   1,        0,        0,        0,\n"
                    "// xy^2,     xyz,      xy,       xz^2,\n"
                    "   0,        1,        0,        0,\n"
                    "// xz,       x,        y^3,      y^2z,\n"
                    "   0,        0,        1,        0,\n"
                    "// y^2,      yz^2,     yz,       y,\n"
                    "   0,        0,        0,        1,\n"
                    "// z^3,      z^2,      z,        C\n"
                    "   0,        0,        0,        1\n"
                    "  >\n"
                    " // sturm // optional, slower but reduces speckles\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-cylinder ()
  (interactive
    (insert (concat
                    "// Capped Cylinder, closed [or open ended]\n"
                    "// cylinder { <END1>, <END2>, RADIUS [open] }\n"
                    "//  END1 = coord of one end of cylinder\n"
                    "//  END2 = coord of other end\n"
                    "// RADIUS = size of cylinder\n"
                    "// open = if present, cylinder is hollow, else capped\n"
                    "cylinder {\n"
                    "  0*x,  3*x,  1\n"
                    "  // open\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-disc ()
  (interactive
    (insert (concat
                    "// flat circular FINITE (no CSG) shape, center hole cutout is optional\n"
                    "disc {\n"
                    "  <0, 1, 0>  // center position\n"
                    "  z,         // normal vector\n"
                    "  1.0,       // outer radius\n"
                    "  0.2        // optional hole radius\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-fractal ()
  (interactive
    (insert (concat
                    "// create a 3-D slice of a 4-D julia fractal object\n"
                    "julia_fractal {\n"
                    "  <-0.083,0.0,-0.83,-0.025>\n"
                    "  quaternion // hypercomplex\n"
                    "  cube // TYPE: sqr | cube | exp | reciprocal | sin | asin | sinh | asinh | cos | acos |\n"
                    "       //       cosh | acosh | tan | atan | tanh | atanh | log | pwr( X_Val, Y_Val )\n"
                    "  max_iteration 8\n"
                    "  precision 20\n"
                    "  // slice <0,0,0,1>,0\n"
                    " }\n"
             )
    )
  )
)

(defun pov-im-height_field+imagemap ()
  (interactive
    (insert (concat
                    "// uses image color index as height, extends along X-Z axes\n"
                    "// from <0 0 0> to <1 1 1>\n"
                    "height_field {\n"
                    "  png \"plasma3.png\"\n"
                    "  texture {\n"
                    "    pigment {\n"
                    "      image_map { png \"plasma3.png\" map_type 0 interpolate 2 once }\n"
                    "      rotate x*90 // lay X-Y image map down onto X-Z height field plane\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-height_field ()
  (interactive
    (insert (concat
                    "// uses image color index as height, extends along X-Z axes\n"
                    "// from <0 0 0> to <1 1 1>\n"
                    "height_field {\n"
                    "  png               // the file type to read (tga/pot/pgm/ppm/png/sys)\n"
                    "  \"plasma3.png\"     // the file name to read\n"
                    "//  [smooth]        // smooth surface normal\n"
                    "//  [water_level N] // truncate/clip below N (0.0 ... 1.0)\n"
                    "//  texture {...}\n"
                    "//  translate VECTOR | rotate VECTOR | scale VECTOR\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-isosurface ()
  (interactive
    (insert (concat
                    "// create a isosurface object - the equipotential surface\n"
                    "// of a 3D math function f(x, y, z)\n"
                    "#declare fn_X = function(x,y,z) { x*x + y*y - 1 }  // cylinder function\n"
                    "isosurface {\n"
                    "  function { x*x + y*y - 1 }          // function (can also contain declared functions\n"
                    "  //function { fn_X(x, y, z) }        // alternative declared function\n"
                    "  contained_by { box { -1.2, 1.2 } }  // container shape\n"
                    "  //threshold 0.0                     // optional threshold value for isosurface [0.0]\n"
                    "  accuracy 0.001                      // accuracy of calculation [0.001]\n"
                    "  max_gradient 4                      // maximum gradient the function can have [1.1]\n"
                    "  //evaluate 5, 1.2, 0.95             // evaluate the maximum gradient\n"
                    "  //max_trace 1                       // maybe increase for use in CSG [1]\n"
                    "  //all_intersections                 // alternative to 'max_trace'\n"
                    "  //open                              // remove visible container surface\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-lathe ()
  (interactive
    (insert (concat
                    "// rotate a 2-D outline of points around the Y axis to create a 3-D shape\n"
                    "lathe {\n"
                    "  linear_spline // linear_spline | quadratic_spline | cubic_spline\n"
                    "  5, // number of points\n"
                    "  <2, 0>, <3, 0>, <3, 5>, <2, 5>, <2, 0> // the list of <u,v> points\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-mesh ()
  (interactive
    (insert (concat
                    "// triangle or smooth-triangle mesh FINITE (no CSG) shape\n"
                    "// NOTE: Each triangle can be independently textured,\n"
                    "//       remaining triangles get texture at bottom\n"
                    "//       For the individual triangles you have to use declared textures\n"
                    "\n"
                    "#declare T1=texture { pigment { color rgb x } } \n"
                    "#declare T2=texture { pigment { color rgb y } } \n"
                    "#declare T3=texture { pigment { color rgb z } } \n"
                    "\n"
                    "mesh { // box example here \n"
                    "  /* top side */\n"
                    "  triangle { <-2,  2, -2>, < 2,  2, -2>, < 2,  2,  2> }\n"
                    "  triangle { <-2,  2, -2>, <-2,  2,  2>, < 2,  2,  2> texture { T1 } }\n"
                    "  /* bottom side */\n"
                    "  triangle { <-2, -2, -2>, < 2, -2, -2>, < 2, -2,  2> }\n"
                    "  triangle { <-2, -2, -2>, <-2, -2,  2>, < 2, -2,  2> }\n"
                    "  /* left side */\n"
                    "  triangle { <-2, -2, -2>, <-2, -2,  2>, <-2,  2,  2> }\n"
                    "  triangle { <-2, -2, -2>, <-2,  2, -2>, <-2,  2,  2> }\n"
                    "  /* right side */\n"
                    "  triangle { < 2, -2, -2>, < 2, -2,  2>, < 2,  2,  2> texture { T2 } }\n"
                    "  triangle { < 2, -2, -2>, < 2,  2, -2>, < 2,  2,  2> texture { T2 } }\n"
                    "  /* front side */\n"
                    "  triangle { <-2, -2, -2>, < 2, -2, -2>, <-2,  2, -2> texture { T3 } }\n"
                    "  triangle { <-2,  2, -2>, < 2,  2, -2>, < 2, -2, -2> texture { T3 } }\n"
                    "  /* back side */\n"
                    "  triangle { <-2, -2,  2>, < 2, -2,  2>, <-2,  2,  2> }\n"
                    "  triangle { <-2,  2,  2>, < 2,  2,  2>, < 2, -2,  2> }\n"
                    "  texture\n"
                    "  { // remaining triangles get this texture\n"
                    "    pigment { color rgb<0.9, 0.9, 0.9> }\n"
                    "    finish { ambient 0.2 diffuse 0.7 }\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-mesh2 ()
  (interactive
    (insert (concat
                    "// new more compact mesh FINITE (no CSG) shape\n"
                    "mesh2 {\n"
                    "  vertex_vectors {\n"
                    "    number_of_vertices,\n"
                    "    <vertex1>, <vertex2>, ...\n"
                    "  }\n"
                    "  normal_vectors {\n"
                    "    number_of_normals,\n"
                    "    <normal1>, <normal2>, ...\n"
                    "  }\n"
                    "  uv_vectors {\n"
                    "    number_of_uv_vectors,\n"
                    "    <uv_vect1>, <uv_vect2>, ...\n"
                    "  }\n"
                    "  texture_list {\n"
                    "    number_of_textures,\n"
                    "    texture { Texture1 },\n"
                    "    texture { Texture2 }, ...\n"
                    "  }\n"
                    "  face_indices {\n"
                    "    number_of_faces,\n"
                    "    <index_a, index_b, index_c> [,texture_index [, texture_index, texture_index]],\n"
                    "    <index_d, index_e, index_f> [,texture_index [, texture_index, texture_index]],\n"
                    "    ...\n"
                    "  }\n"
                    "  normal_indices {\n"
                    "    number_of_faces,\n"
                    "    <index_a, index_b, index_c>,\n"
                    "    <index_d, index_e, index_f>,\n"
                    "    ...\n"
                    "  }\n"
                    "  uv_indices {\n"
                    "    number_of_faces,\n"
                    "    <index_a, index_b, index_c>,\n"
                    "    <index_d, index_e, index_f>,\n"
                    "    ...\n"
                    "  }\n"
                    "  [object modifiers]\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-object ()
  (interactive
    (insert (concat
                    "// general purpose enclosing wrapper for a predefined shape\n"
                    "#declare MyShape = box {-1,1}    // declare the shape\n"
                    "object { MyShape }               // create an actual object with that shape\n"
                    "object { MyShape translate 3*x}  // create another object with that shape\n"
             )
    )
  )
)

(defun pov-im-parametric ()
  (interactive
    (insert (concat
                    "// parametric object\n"
                    "// not yet worked out\n"
                    "parametric {\n"
                    "  function { u*v*sin (15*v) },            // x-axis\n"
                    "  function { v },                         // y-axis\n"
                    "  function { u*v*cos (15*v) }             // z-axis\n"
                    "  <0,0>, <1,1>\n"
                    "  // contained_by { sphere { 0,1 } } // texturing problem, use box instead to see!\n"
                    "  contained_by { box { <-1,-1,-1>, <1,1,1> } }\n"
                    "  // max_gradient 2\n"
                    "  accuracy 0.005        // 0.001 default, lower slower but better\n"
                    "  precompute 15 x,y,z   // precompute [x,y,z] normally gives faster rendering (<=20)\n"
                    "\n"
                    "  rotate 180*x\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-plane ()
  (interactive
    (insert (concat
                    "// An infinite planar surface\n"
                    "// plane {<A, B, C>, D } where: A*x + B*y + C*z = D\n"
                    "plane {\n"
                    "  y, // <X Y Z> unit surface normal, vector points \"away from surface\"\n"
                    "  -1.0 // distance from the origin in the direction of the surface normal\n"
                    "  hollow on // has an inside pigment?\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-polygon ()
  (interactive
    (insert (concat
                    "// arbitrary X,Y FINITE (no CSG) shape\n"
                    "// A complex example for a polygon is the letter \"P\":\n"
                    "polygon {\n"
                    "  12, // number of points\n"
                    "  <0, 0>, <0, 6>, <4, 6>, <4, 3>, <1, 3>, <1, 0>, <0, 0>, // list of <u,v> points\n"
                    "  <1, 4>, <1, 5>, <3, 5>, <3, 4>, <1, 4>\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-polynomial ()
  (interactive
    (insert (concat
                    "// create an Nth order infinite polynomial surface\n"
                    "// poly { N <a,b,c...> [sturm] }\n"
                    "// N = order of poly, M terms where M = (N+1)*(N+2)*(N+3)/6\n"
                    "poly {\n"
                    "  5, // order of polynomial (2...7)\n"
                    "  <\n"
                    "// x^5,        x^4y,       x^4z,       x^4,\n"
                    "   0,          0,          0,          0,\n"
                    "// x^3y^2,     x^3yz,      x^3y,       x^3z^2,\n"
                    "   0,          0,          0,          0,\n"
                    "// x^3z,       x^3,        x^2y^3,     x^2y^2z,\n"
                    "   0,          0,          0,          0,\n"
                    "// x^2y^2,     x^2yz^2,    x^2yz,      x^2y,\n"
                    "   0,          0,          0,          0,\n"
                    "// x^2z^3,     x^2z^2,     x^2z,       x^2,\n"
                    "   0,          0,          0,          0,\n"
                    "// xy^4,       xy^3z,      xy^3,       xy^2z^2,\n"
                    "   0,          0,          0,          0,\n"
                    "// xy^2z,      xy^2,       xyz^3,      xyz^2,\n"
                    "   0,          0,          0,          0,\n"
                    "// xyz,        xy,         xz^4,       xz^3,\n"
                    "   0,          0,          0,          0,\n"
                    "// xz^2,       xz,         x,          y^5,\n"
                    "   0,          0,          0,          0,\n"
                    "// y^4z,       y^4,        y^3z^2,     y^3z,\n"
                    "   0,          0,          0,          0,\n"
                    "// y^3,        y^2z^3,     y^2z^2,     y^2z,\n"
                    "   0,          0,          0,          0,\n"
                    "// y^2,        yz^4,       yz^3,       yz^2,\n"
                    "   0,          0,          0,          0,\n"
                    "// yz,         y,          z^5,        z^4,\n"
                    "   0,          0,          0,          0,\n"
                    "// z^3,        z^2,        z,          C           \n"
                    "   0,          0,          0,          0\n"
                    "  >\n"
                    "  sturm // optional, slower but reduces speckles\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-prism ()
  (interactive
    (insert (concat
                    "// extrude a closed 2-D shape along an axis\n"
                    "prism {                                \n"
                    "  linear_sweep  // or conic_sweep for tapering to a point\n"
                    "  linear_spline // linear_spline | quadratic_spline | cubic_spline | bezier_spline \n"
                    "  -0.5,         // height 1\n"
                    "   0.5,         // height 2\n"
                    "  10,           // number of points\n"
                    "  // (--- the <u,v> points ---)\n"
                    "  < 0.2, -1.0>, < 0.2,  0.2>, < 1.0, -0.2>, < 1.0,  0.2>, < 0.2,  1.0>, \n"
                    "  <-0.2,  1.0>, <-1.0,  0.2>, <-1.0, -0.2>, <-0.2,  0.2>, <-0.2, -1.0>\n"
                    "  // , <0.2,   0.2> // match 2nd point, if quadratic_spline add this\n"
                    "  // , <0.2,  -1.0> // match 1st point, if cubic_spline add this as well as the other\n"
                    "  // [open]\n"
                    "  // [sturm]\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-quadric ()
  (interactive
    (insert (concat
                    "// create a quadratic (2nd order) infinite polynomial surface\n"
                    "quadric {\n"
                    "  <0, 1, 1> //  A x^2  + B y^2  + C z^2  +\n"
                    "  <1, 0, 0> //  D xy   + E xz   + F yz   +\n"
                    "  <0, 3, 0> //  G x    + H y    + I z    +\n"
                    "  2         //  J\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-quartic ()
  (interactive
    (insert (concat
                    "// create a 4th order infinite polynomial surface\n"
                    "quartic {\n"
                    "<\n"
                    "// x^4,        x^3y,       x^3z,       x^3,        x^2y^2,\n"
                    "   0,          0,          0,          0,          0,\n"
                    "// x^2yz,      x^2y,       x^2z^2,     x^2z,       x^2,\n"
                    "   0,          0,          0,          0,          0,\n"
                    "// xy^3,       xy^2z,      xy^2,       xyz^2,      xyz,\n"
                    "   0,          0,          0,          0,          0,\n"
                    "// xy,         xz^3,       xz^2,       xz,         x,\n"
                    "   0,          0,          0,          0,          0,\n"
                    "// y^4,        y^3z,       y^3,        y^2z^2,     y^2z,\n"
                    "   0,          0,          0,          0,          0,\n"
                    "// y^2,        yz^3,       yz^2,       yz,         y,\n"
                    "   0,          0,          0,          0,          0,\n"
                    "// z^4,        z^3,        z^2,        z,          C           \n"
                    "   0,          0,          0,          0,          0\n"
                    ">\n"
                    "  sturm // optional, slower but reduces speckles\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-smooth_triangle ()
  (interactive
    (insert (concat
                    "// rounded-surface FINITE (no CSG) triangle shape\n"
                    "smooth_triangle {\n"
                    "  <  0,  30, 0> <0,  0.7071, -0.7071>  // <Vertex1>  <SurfaceNormal1>\n"
                    "  < 40, -20, 0> <0, -0.8664, -0.5>     // <Vertex2>  <SurfaceNormal2>\n"
                    "  <  0,   0, 0> <0, -0.5,    -0.8664>  // <Vertex3>  <SurfaceNormal3>\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-sphere ()
  (interactive
    (insert (concat
                    "// create a sphere shape\n"
                    "sphere {\n"
                    "  <0, 1, 0> // center of sphere <X Y Z>\n"
                    "  0.5       // radius of sphere\n"
                    "  // scale <1,2,1> // <= Note: Spheres can become ellipses by uneven scaling\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-sphere_sweep ()
  (interactive
    (insert (concat
                    "// create a curved tube object translating a sphere along a certain path\n"
                    "sphere_sweep {\n"
                    "  linear_spline                 // linear curve\n"
                    "  //cubic_spline                // alternative spline curves\n"
                    "  //b_spline\n"
                    "  4,                            // number of specified sphere positions\n"
                    "  <-5, -5, 0>, 1                // position, radius\n"
                    "  <-5, 5, 0>, 1                 // ...\n"
                    "  < 5, -5, 0>, 1\n"
                    "  < 5, 5, 0>, 1\n"
                    "  //tolerance 0.001             // optional\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-superellipsoid ()
  (interactive
    (insert (concat
                    "// create a superquadric ellipsoid shape\n"
                    "// As the exponents approach 1.0, the edges get rounder\n"
                    "superellipsoid {\n"
                    "  <0.5,  // east-west exponent (0.0 ... 1.0)\n"
                    "  1.0>   // north-south exponent (0.0 ... 1.0)\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-surface-of-revolution ()
  (interactive
    (insert (concat
                    "// create a Surface of Revolution shape (like lathe, but faster)\n"
                    "sor {\n"
                    "    7, // # of points\n"
                    "    <0.000000, 0.000000> // list of <u,v> points\n"
                    "    <0.118143, 0.000000>\n"
                    "    <0.620253, 0.540084>\n"
                    "    <0.210970, 0.827004>\n"
                    "    <0.194093, 0.962025>\n"
                    "    <0.286920, 1.000000>\n"
                    "    <0.468354, 1.033755>\n"
                    "    // [open]\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-text ()
  (interactive
    (insert (concat
                    "// create a TrueType text shape\n"
                    "text {\n"
                    "  ttf             // font type (only TrueType format for now)\n"
                    "  \"crystal.ttf\",  // Microsoft Windows-format TrueType font file name\n"
                    "  \"POV-Ray\",      // the string to create\n"
                    "  2,              // the extrusion depth\n"
                    "  0               // inter-character spacing\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-torus ()
  (interactive
    (insert (concat
                    "// torus {MAJOR, MINOR} // (in the X-Z plane)\n"
                    "// MAJOR = float value giving the major radius\n"
                    "// MINOR = float specifying the minor radius\n"
                    "// The major radius extends from the center of the hole\n"
                    "// to the mid-line of the rim while the minor radius\n"
                    "// is the radius of the cross-section of the rim.\n"
                    "torus {\n"
                    "  0.8,\n"
                    "  0.2\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-triangle ()
  (interactive
    (insert (concat
                    "// triangular FINITE (no CSG) shape (vertices are endpoints)\n"
                    "triangle {\n"
                    "  <-1,  0, -1>,  // <Vertex1>\n"
                    "  < 1,  0, -1>,  // <Vertex2>\n"
                    "  < 0,  0,  1>   // <Vertex3>\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-box ()
  (interactive
    (insert (concat
                    "box { <-1, -1, -1>, <1, 1, 1> }\n"
             )
    )
  )
)

(defun pov-im-cone ()
  (interactive
    (insert (concat
                    "cone { <0, 1, 0>, 0.0, <0, -1, 0>, 1.0 }\n"
             )
    )
  )
)

(defun pov-im-cylinder ()
  (interactive
    (insert (concat
                    "cylinder { <0, 0, 0>, <3, 0, 0>, 1 }\n"
             )
    )
  )
)

(defun pov-im-disc ()
  (interactive
    (insert (concat
                    "disc { <0, 0, 0>, z, 1.0, 0.2 }\n"
             )
    )
  )
)

(defun pov-im-plane ()
  (interactive
    (insert (concat
                    "plane { y, -1.0 }\n"
             )
    )
  )
)

(defun pov-im-sphere ()
  (interactive
    (insert (concat
                    "sphere { <0, 1, 0>, 0.5 }\n"
             )
    )
  )
)

(defun pov-im-superellipsoid ()
  (interactive
    (insert (concat
                    "superellipsoid { <0.5, 1.0> }\n"
             )
    )
  )
)

(defun pov-im-torus ()
  (interactive
    (insert (concat
                    "torus { 0.8, 0.2 }\n"
             )
    )
  )
)

(defun pov-im-bounded_by ()
  (interactive
    (insert (concat
                    "// set an outer proposed boundary for parent object(s)\n"
                    "// allows the ray-tracer to do a quick check on the Bounding Shape,\n"
                    "// and if outside, it skips checking this object, which can speed\n"
                    "// up overall rendering of complex shapes.  Note that this\n"
                    "// is not a 100% guaranteed clipping shape... use the\n"
                    "// clipped_by statement for that.\n"
                    "bounded_by { box { -(x+y+z) +(x+y+z) }\n"
             )
    )
  )
)

(defun pov-im-clipped_by ()
  (interactive
    (insert (concat
                    "// set clipping shape for parent object.\n"
                    "// (parent shape will not extend beyond ClipShape bounds)\n"
                    "clipped_by { sphere { 0*x, 1.0 } }\n"
             )
    )
  )
)

(defun pov-im-double_illuminate ()
  (interactive
    (insert (concat
                    "// illuminate also the shadow side of a surface\n"
                    "  double_illuminate\n"
             )
    )
  )
)

(defun pov-im-hollow ()
  (interactive
    (insert (concat
                    "// allow atmosphere inside shape\n"
                    "  hollow\n"
             )
    )
  )
)

(defun pov-im-no_image ()
  (interactive
    (insert (concat
                    "// show reflections, but not the object directly\n"
                    "  no_image\n"
             )
    )
  )
)

(defun pov-im-no_reflection ()
  (interactive
    (insert (concat
                    "// show no reflections of the object\n"
                    "  no_reflection\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-no_shadow ()
  (interactive
    (insert (concat
                    "// give object no shadow\n"
                    "  no_shadow\n"
             )
    )
  )
)

(defun pov-im-open ()
  (interactive
    (insert (concat
                    "// some objects (cylinder, cone, prism, etc.) can have their ends open\n"
                    "  open\n"
             )
    )
  )
)

(defun pov-im-sturm ()
  (interactive
    (insert (concat
                    "// use slower but more accurate root solver.\n"
                    "// can be applied to:\n"
                    "//    -blob\n"
                    "//    -cubic\n"
                    "//    -lathe (only with quadratic splines)\n"
                    "//    -poly\n"
                    "//    -prism (only with cubic splines)\n"
                    "//    -quartic\n"
                    "//    -sor\n"
                    "  sturm\n"
             )
    )
  )
)

(defun pov-im-cutaway_textures ()
  (interactive
    (insert (concat
                    "// cutaway_textures can be useful for texturing differences\n"
                    "difference {\n"
                    "  sphere {\n"
                    "    0, 1\n"
                    "    texture { pigment { color Red } }\n"
                    "  }\n"
                    "  box { 0, 1 }\n"
                    "  // the visible parts of this object will get the texture of the sphere\n"
                    "  cutaway_textures\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-difference ()
  (interactive
    (insert (concat
                    "// CSG difference, subtract intersections of shapes 2...N from Shape1\n"
                    "difference {\n"
                    "  Shape1 {...} // Start with this shape\n"
                    "  Shape2 {...} // This will be \"cut out\" of Shape1\n"
                    "  ShapeN {...} // This will be \"cut out\" of Shape1\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-intersection ()
  (interactive
    (insert (concat
                    "// CSG intersection, the common space where all the shapes meet\n"
                    "intersection {\n"
                    "  Shape1 {...}\n"
                    "  Shape2 {...}\n"
                    "  ShapeN {...}\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-inverse ()
  (interactive
    (insert (concat
                    "// flip an object's inside and outside for intersect/difference\n"
                    "intersection {\n"
                    "  Shape1 {... inverse }\n"
                    "  Shape2 {...}\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-merge ()
  (interactive
    (insert (concat
                    "// CSG merge, merge all of shapes 1...N\n"
                    "// like 'union', but melted together so no overlap seam lines inside\n"
                    "merge {\n"
                    "  Shape1 {...}\n"
                    "  Shape2 {...}\n"
                    "  ShapeN {...}\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-union ()
  (interactive
    (insert (concat
                    "// CSG union, add all of shapes 1...N\n"
                    "union {\n"
                    "  Shape1 {...}\n"
                    "  Shape2 {...}\n"
                    "  ShapeN {...}\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-interior ()
  (interactive
    (insert (concat
                    "// describes inside of a shape (similar to texture, but for inside, not surface)\n"
                    "  interior {\n"
                    "    ior 1.31              // index of refraction\n"
                    "    dispersion            // needs a color_map in the light source\n"
                    "    dispersion_samples 7  // accuracy of calculation (2..100) [7]\n"
                    "    caustics 1.0          // faked caustics (0..1)\n"
                    "    fade_power 1001       // values larger than 1000 give realistic exponential attenuation\n"
                    "    fade_distance 3       // distance where light reaches half intensity\n"
                    "    fade_color <0, .3, 0> // color for fading\n"
                    "    media {..media items..}\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-material ()
  (interactive
    (insert (concat
                    "// material {} is a wrapper for interior and texture\n"
                    "// you can declare it like texture and interior too\n"
                    "// Note that it has nothing to do with material_map\n"
                    "material {\n"
                    "  texture {\n"
                    "    pigment { color rgbt <0.5, 0.5, 0.7, 0.5> }\n"
                    "    finish { diffuse 0.2 specular 0.6 }\n"
                    "  }\n"
                    "  interior {\n"
                    "    ior 1.5\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-media ()
  (interactive
    (insert (concat
                    "// general media statement\n"
                    "// put in interior or in scene file for atmospheric media\n"
                    "media {\n"
                    "  // (---general values---)\n"
                    "  intervals 10           // number of intervals used for sampling [10]\n"
                    "  samples 1,1            // minimum and maximum number of samples taken per interval [1,1]\n"
                    "  confidence 0.9         // statistic parameter higher->better quality [0.9]\n"
                    "  variance 1.0/128       // statistic parameter lower->better quality [1.0/128]\n"
                    "  ratio 0.9              // distribution between lit and unlit areas [0.9]\n"
                    "  // (---media types---)\n"
                    "  absorption rgb<0,1,0>  // absorbing media, block light of specified color\n"
                    "  //emission rgb<0,1,0>  // emitting media, emit light of specified color\n"
                    "  /*\n"
                    "  scattering {           // scattering media\n"
                    "    1,                   // scattering type 1=isotropic; 2=Mie haze; 3=Mie murky\n"
                    "                         // 4=Rayleigh; 5=Henyey-Greenstein\n"
                    "    rgb<0,1,0>           // color\n"
                    "    //eccentricity 0.25  // eccentricity for type 5 [0.0]\n"
                    "    //extinction 1.0     // for balancing amount of absorption [1.0]\n"
                    "  }*/\n"
                    "  // (---method---)\n"
                    "  //method 1             // old Povray 3.1 method\n"
                    "  //method 2             // new method with even distribution\n"
                    "  method 3               // adaptive sampling\n"
                    "  //aa_threshold 0.1     // accuracy threshold for method 3 [0.1]\n"
                    "  //aa_level 4           // maximum recursion depth for method 3 [4]\n"
                    "  //jitter 0.5           // randomness for method 2&3\n"
                    "  // (---density---)\n"
                    "  /*\n"
                    "  density {\n"
                    "    spherical            // any pattern\n"
                    "    color_map {...}      // color_map\n"
                    "    //density_map {...}  // or alternatively density_map\n"
                    "  }*/\n"
                    "\n"
                    "  // translate etc.\n"
                    "}\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-bump_map-normal ()
  (interactive
    (insert (concat
                    "// texture normal {} attribute\n"
                    "// create a texture that has a bumpiness corresponding to color index\n"
                    "// image maps into X-Y plane from <0,0,0> to <1,1,0>\n"
                    "bump_map { // uses image color or index as bumpiness\n"
                    "  png \"plasma3.png\" // the file to read (iff/tga/gif/png/jpeg/tiff/sys)\n"
                    "  map_type 0        // 0=planar, 1=spherical, 2=cylindrical, 5=torus\n"
                    "  interpolate 2     // 0=none, 1=linear, 2=bilinear, 4=normalized distance\n"
                    "  // once           // for no repetitive tiling\n"
                    "  // [use_color | use_index]\n"
                    "  bump_size 0.5     // 0...3\n"
                    "} // bump_map\n"
             )
    )
  )
)

(defun pov-im-image_map-pigment ()
  (interactive
    (insert (concat
                    "// texture pigment {} attribute\n"
                    "// create a texture that lays an image's colors onto a surface\n"
                    "// image maps into X-Y plane from <0,0,0> to <1,1,0>\n"
                    "image_map {\n"
                    "  png \"plasma3.png\" // the file to read (iff/tga/gif/png/jpeg/tiff/sys)\n"
                    "  map_type 0        // 0=planar, 1=spherical, 2=cylindrical, 5=torus\n"
                    "  interpolate 2     // 0=none, 1=linear, 2=bilinear, 4=normalized distance\n"
                    "  // once           // for no repetitive tiling\n"
                    "  // [filter N V]   // N=all or color index # (0...N), V= value (0.0...1.0)\n"
                    "  // [transmit N V] // N=all or color index # (0...N), V= value (0.0...1.0)\n"
                    "  // [use_color | use_index]\n"
                    "} // image_map\n"
             )
    )
  )
)

(defun pov-im-material_map-texture ()
  (interactive
    (insert (concat
                    "// texture attribute\n"
                    "// create a texture that maps different textures onto different color indexes\n"
                    "// image maps into X-Y plane from <0,0,0> to <1,1,0>\n"
                    "material_map {\n"
                    "  png \"plasma3.png\" // the file to read (iff/tga/gif/png/jpeg/tiff/sys)\n"
                    "  map_type 0        // 0=planar, 1=spherical, 2=cylindrical, 5=torus\n"
                    "  interpolate 2     // 0=none, 1=linear, 2=bilinear, 4=normalized distance\n"
                    "  // once           // for no repetitive tiling\n"
                    "  texture { pigment {wood} }     // texture for color index # 0\n"
                    "  texture { pigment {granite} }  // texture for color index # 1\n"
                    "  texture { pigment {agate} }    // texture for color index # 2\n"
                    "  // etc...\n"
                    "} // material_map\n"
             )
    )
  )
)

(defun pov-im-tiles-texture ()
  (interactive
    (insert (concat
                    "// texture component\n"
                    "// create a tiled texture, alternates between 2 full textures\n"
                    "tiles {\n"
                    "  texture {   // full texture of tile #1\n"
                    "    pigment { agate scale 0.3 }\n"
                    "    finish { ambient 0.2 }\n"
                    "  }\n"
                    "  tile2\n"
                    "  texture {   // full texture of tile #2\n"
                    "    pigment { granite }\n"
                    "    finish { specular 0.3 reflection 0.2 }\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-color ()
  (interactive
    (insert (concat
                    "// texture pigment attribute\n"
                    "// RGB values can range from 0.0 (dark) to 1.0 (saturated)\n"
                    "  color red 0.5 green 0.5 blue 0.5 // filter 0.8\n"
                    "// (---alternative form---)\n"
                    "  //color rgb <0.5, 0.5, 0.5>\n"
                    "  //color rgbf <0.5, 0.5, 0.5, 0.8>\n"
             )
    )
  )
)

(defun pov-im-color_map ()
  (interactive
    (insert (concat
                    "// texture pigment {} attribute\n"
                    "// create a smooth color gradation map\n"
                    "color_map {\n"
                    "  //  [ END_VAL color red R green G blue B filter F transmit T]\n"
                    "  [ 0.1  color red 0.0 green 0.0 blue 0.0]\n"
                    "  [ 0.3  color red 0.8 green 0.4 blue 0.2]\n"
                    "  [ 0.5  color red 0.0 green 0.0 blue 0.0]\n"
                    "  [ 0.7  color red 0.0 green 0.0 blue 0.8 filter 0.5]\n"
                    "  [ 1.0  color red 0.8 green 0.8 blue 0.8 filter 1.0]\n"
                    "  // (---alternative form---)\n"
                    "  //[ 1.0  color rgbf <0.8, 0.8, 0.8, 1.0> ]\n"
                    "} // color_map\n"
             )
    )
  )
)

(defun pov-im-finish-properties ()
  (interactive
    (insert (concat
                    "// control an object's surface finish\n"
                    "  finish {\n"
                    "    ambient 0.1          // ambient surface reflection color [0.1]\n"
                    "    // (---diffuse lighting---)\n"
                    "    diffuse 0.6          // amount [0.6]\n"
                    "    brilliance 1.0       // tightness of diffuse illumination [1.0]\n"
                    "    // (---phong highlight---)\n"
                    "    //phong 0.5          // amount [0.0]\n"
                    "    //phong_size 40      // (1.0..250+) (dull->highly polished) [40]\n"
                    "    // (---specular highlight---)\n"
                    "    //specular 0.5       // amount [0.0]\n"
                    "    //roughness 0.05     // (~1.0..0.0005) (dull->highly polished) [0.05]\n"
                    "    // (---phong and specular---)\n"
                    "    //metallic [Amount]  // give highlight color of surface\n"
                    "    // (---reflection---)\n"
                    "    /*\n"
                    "    reflection {\n"
                    "      //0.0                      // minimal reflection value (for variable reflection)\n"
                    "      1.0                        // reflection amount (maximum for variable reflection)\n"
                    "      //fresnel on               // realistic variable reflection\n"
                    "      //falloff 1.0              // falloff exponent for variable reflection\n"
                    "      //exponent 1.0             // influence surface reflection characteristic\n"
                    "      //metallic 1.0             // tint reflection in surface color\n"
                    "    }*/\n"
                    "    //conserve_energy            // more realistic\n"
                    "    // (---others---)\n"
                    "    //crand 0.2                  // randomly speckle the surface [0.0]\n"
                    "    /*\n"
                    "    irid {               // Iridescence (Newton's thin film interference)\n"
                    "      0.25               // intensity\n"
                    "      //thickness 0.0    // film's thickness [0.0]\n"
                    "      //turbulence 0.3   // film's thickness turbulence\n"
                    "    }*/\n"
                    "  } // finish\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-function-image-type ()
  (interactive
    (insert (concat
                    "// use this for example in a heightfield or an image_map\n"
                    "  function\n"
                    "  50, 50 {            // resolution of generated image\n"
                    "    pattern {         // user defined function\n"
                    "      agate            \n"
                    "    }\n"
                    "  }\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-interior_texture ()
  (interactive
    (insert (concat
                    "// with interior_texture you can give the interior surface\n"
                    "// of an object a different texture\n"
                    "object {\n"
                    "  sphere { 0, 1 }\n"
                    "  texture { pigment { color Red } }\n"
                    "  interior_texture { pigment { color Blue } }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-normal-component ()
  (interactive
    (insert (concat
                    "// texture component\n"
                    "  normal {\n"
                    "    bumps 0.5         // any pattern optionally followed by an intensity value [0.5]\n"
                    "    //bump_size 2.0   // optional\n"
                    "    //accuracy 0.02   // changes the scale for normal calculation [0.02]\n"
                    "    //scale 0.3       // any transformations\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-pigment-component ()
  (interactive
    (insert (concat
                    "// texture component\n"
                    "  pigment {\n"
                    "    color rgb <0.7,0.5,0.3>     // solid color pigment\n"
                    "    /*\n"
                    "    agate                       // patterned pigment\n"
                    "    color_map {\n"
                    "      [0.0 color Green ]\n"
                    "      [1.0 color Blue ]\n"
                    "    } */\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-quick_color-attribute ()
  (interactive
    (insert (concat
                    "// texture pigment attribute\n"
                    "quick_color green 1 // low-quality quick-render color\n"
             )
    )
  )
)

(defun pov-im-texture-example ()
  (interactive
    (insert (concat
                    "texture {\n"
                    "  pigment {          // (---surface color---)\n"
                    "    marble           // some pattern\n"
                    "    color_map {      // color map\n"
                    "      [0.1 color red 1]\n"
                    "      [0.5 color rgbf 1]\n"
                    "    }\n"
                    "    turbulence 0.5   // some turbulence\n"
                    "    scale <1,3,1>    // transformations\n"
                    "  }\n"
                    "  normal {           // (---surface bumpiness---)\n"
                    "    marble 0.3       // some pattern (and intensity)\n"
                    "    turbulence 0.5   // some turbulence\n"
                    "    scale <1,3,1>    // transformations\n"
                    "  }\n"
                    "  finish {           // (---surface finish---)\n"
                    "    ambient 0.2\n"
                    "    specular 0.6     // shiny\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-triangle-texture-interpolation ()
  (interactive
    (insert (concat
                    "// interpolate the triangle's texture between 3\n"
                    "// specified textures for the 3 vertices\n"
                    "mesh {\n"
                    "  triangle {\n"
                    "    <-1,  0, -1>, < 1,  0, -1>, < 0,  0,  1>    // vertices\n"
                    "    texture_list {\n"
                    "      Texture_01           // 3 previously declared textures\n"
                    "      Texture_02\n"
                    "      Texture_03\n"
                    "    }\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-uv_mapping ()
  (interactive
    (insert (concat
                    "// sample for uv_mapping with a bicubic patch\n"
                    "// uv_mapping can also be used for:\n"
                    "// - mesh, mesh2\n"
                    "// - lathe\n"
                    "// - sor\n"
                    "// - sphere\n"
                    "// - box\n"
                    "bicubic_patch {\n"
                    "  type 0\n"
                    "  flatness 0.01\n"
                    "  u_steps 4\n"
                    "  v_steps 4\n"
                    "  <0, 0, 2>, <1, 0, 0>, <2, 0, 0>, <3, 0, -2>,\n"
                    "  <0, 1  0>, <1, 1, 0>, <2, 1, 0>, <3, 1,  0>,\n"
                    "  <0, 2, 0>, <1, 2, 0>, <2, 2, 0>, <3, 2,  0>,\n"
                    "  <0, 3, 2>, <1, 3, 0>, <2, 3, 0>, <3, 3, -2>\n"
                    "\n"
                    "  uv_mapping\n"
                    "\n"
                    "  texture {\n"
                    "    pigment {\n"
                    "      checker\n"
                    "      color Red\n"
                    "      color Blue\n"
                    "      scale 0.1\n"
                    "    }\n"
                    "    finish { specular 0.5 }\n"
                    "  }\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-agate ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  agate            // use an agate-like texture\n"
                    "  //agate 0.5      // for use with normal{} (0...1 or more)\n"
                    "  agate_turb 0.3   // can alter turbulence [1.0]\n"
             )
    )
  )
)

(defun pov-im-average ()
  (interactive
    (insert (concat
                    "// average pattern averaging several pigments\n"
                    "// can also be used for texture_map, density_map, or normal_map\n"
                    "  pigment {\n"
                    "    average\n"
                    "    pigment_map {\n"
                    "      [1.0  Pigment_1]     // weighting 1.0\n"
                    "      [0.5  Pigment_2]     // weighting 0.5\n"
                    "    }\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-boxed ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// originally designed for use with media density\n"
                    "  boxed     // box shaped gradient around origin\n"
             )
    )
  )
)

(defun pov-im-bozo ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  bozo              // very smooth, random noise function\n"
                    "  //bozo 0.5        // for use with normal{} (0...1 or more)\n"
                    "  //turbulence 0.3  // turbulence often useful\n"
             )
    )
  )
)

(defun pov-im-brick ()
  (interactive
    (insert (concat
                    "// solid pattern for use in texture/pigment/normal(/density)\n"
                    "  // brick pattern generates a brick wall like structure\n"
                    "  brick      \n"
                    "  color Gray,          // color of mortar\n"
                    "  color Red            // color of brick\n"
                    "  brick_size <2,1,1>   // size of brick    (optional)\n"
                    "  mortar     0.2       // width of mortar  (optional)\n"
             )
    )
  )
)

(defun pov-im-bumps ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// bumps is identical to bozo when used in pigment/density\n"
                    "// especially suited for normal {}\n"
                    "  bumps 0.5    // bumpy surface (0...1 or more)\n"
                    "  //bumps      // for use with pigment{}\n"
             )
    )
  )
)

(defun pov-im-cells ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  cells            // random value cube cells\n"
                    "  //cells 0.5      // for use with normal{} (0...1 or more)\n"
             )
    )
  )
)

(defun pov-im-checker ()
  (interactive
    (insert (concat
                    "// solid pattern for use in texture/pigment/normal(/density)\n"
                    "// cube checker pattern, alternates color1 and color2\n"
                    "  checker\n"
                    "    color rgb <1, 1, 1>,\n"
                    "    color rgb <0, 1, 0>\n"
             )
    )
  )
)

(defun pov-im-crackle ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "   crackle          // crackled pattern\n"
                    "   //crackle 0.5    // for use with normal{} (0...1 or more)\n"
                    "   // (---additional parameters---)\n"
                    "   //form <-1,1,0>\n"
                    "   //metric 2\n"
                    "   //offset 0.0\n"
                    "   //solid\n"
             )
    )
  )
)

(defun pov-im-cylindrical ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// originally designed for use with media density\n"
                    "  cylindrical   // cylinder shaped gradient along the Y-axis\n"
             )
    )
  )
)

(defun pov-im-density_file ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  density_file df3 \"spiral.df3\"         // 3D raster file (df3) pattern\n"
                    "  //density_file df3 \"spiral.df3\", 0.5  // for use with normal{} (0...1 or more)\n"
                    "  //interpolate 1                       // optional interpolation\n"
             )
    )
  )
)

(defun pov-im-dents ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// especially suited for normal {}\n"
                    "  dents 0.6    // dented surface (0...1 or more)\n"
                    "  //dents      // for use with pigment{}\n"
             )
    )
  )
)

(defun pov-im-facets ()
  (interactive
    (insert (concat
                    "// pattern for use in normal {}\n"
                    "// not usable in pigments\n"
                    "  facets          // generates facets on the surface\n"
                    "  coords 1.0      // for flat surfaces\n"
                    "  //size 1.0      // for curved surfaces\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-fractal ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// various types of fractals\n"
                    "  // (---mandelbrot---)\n"
                    "  mandel 50               // number of iterations\n"
                    "         //2,             // optional exponent (2..4) [2]\n"
                    "  // (---julia---)\n"
                    "  julia <0.5,0.8>,        // complex number (2d vector)\n"
                    "        50                // number of iterations\n"
                    "        //2,              // optional exponent (2..4) [2]\n"
                    "  // (---magnet mandel---)\n"
                    "  magnet 1                // type (1 or 2)\n"
                    "   mandel 50              // number of iterations\n"
                    "  // (---magnet julia---)\n"
                    "  magnet 1                // type (1 or 2)\n"
                    "   julia <0.5,0.8>,       // complex number (2d vector)\n"
                    "         50               // number of iterations\n"
                    "\n"
                    "  //exterior 1, 1         // optional: special handling of exterior\n"
                    "  //interior 0, 1         // optional: special handling of interior\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-function ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// function pattern using a user defined function\n"
                    "  #declare fn_X = function(x,y,z) { z + x*x + y*y }  // parabolid function\n"
                    "  function { fn_X(x, y, z) }\n"
                    "  //function { fn_X(x, y, z) } 0.5  // for use with normal{} (0...1 or more)\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-gradient ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  gradient y         // specify gradient direction vector\n"
                    "  //gradient y, 0.5  // for use with normal{} (0...1 or more)\n"
             )
    )
  )
)

(defun pov-im-granite ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  granite        // create a granite-like texture\n"
                    "  //granite 0.5  // for use with normal{} (0...1 or more)\n"
             )
    )
  )
)

(defun pov-im-hexagon ()
  (interactive
    (insert (concat
                    "// solid pattern for use in texture/pigment/normal(/density)\n"
                    "// create a hexagonal pattern texture with 3 colors\n"
                    "// hexagons are in the X-Z plane, extending infinitely in Y plane\n"
                    "  hexagon\n"
                    "    color blue 1,\n"
                    "    color red 1,\n"
                    "    color green 1\n"
             )
    )
  )
)

(defun pov-im-image_pattern ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// takes the color of an image file to calculate the pattern value\n"
                    "  image_pattern {\n"
                    "    png \"plasma3.png\"   // image file to read (iff/tga/gif/png/jpeg/tiff/sys)\n"
                    "    //use_color         // use pixel color gray scale (default)\n"
                    "    //use_alpha         // use alpha value\n"
                    "    //use_index         // use palette index\n"
                    "  }\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-leopard ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  leopard        // create a leopard spotted texture\n"
                    "  //leopard 0.5  // for use with normal{} (0...1 or more)\n"
             )
    )
  )
)

(defun pov-im-marble ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// similar to gradient, gives a nice marble with turbulence\n"
                    "  marble           // create a marble-like texture\n"
                    "  //marble 0.5     // for use with normal{} (0...1 or more)\n"
                    "  //turbulence 1.0 // should have some!\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-object ()
  (interactive
    (insert (concat
                    "// solid pattern for use in texture/pigment/normal(/density)\n"
                    "  object {        // create a solid pattern based on an object shape\n"
                    "    sphere { <0,0,0>, 1 }\n"
                    "    color rgb <1,0,0>,      // outside object\n"
                    "    color rgb <0,0,1>       // inside object\n"
                    "  }\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-onion ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  onion          // create an onion-like concentric circle texture\n"
                    "  //onion 0.5    // for use with normal{} (0...1 or more)\n"
             )
    )
  )
)

(defun pov-im-pigment_pattern ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// takes the color of an pigment to calculate the pattern value\n"
                    "   pigment_pattern {\n"
                    "     checker White, Black   // can use any pigment here\n"
                    "   }\n"
             )
    )
  )
)

(defun pov-im-planar ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// originally designed for use with media density\n"
                    "  planar         // similar to gradient y, but only +/- 1 unit around origin\n"
             )
    )
  )
)

(defun pov-im-quilted ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// creates a pattern like a quilt\n"
                    "  quilted 0.4 // amount\n"
                    "  control0 -1 // lower curvature (-1...1)\n"
                    "  control1 1  // upper curvature (-1...1)\n"
             )
    )
  )
)

(defun pov-im-radial ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// radially projected color_map in X-Z\n"
                    "  radial\n"
                    "  //radial 0.5     // for use with normal{} (0...1 or more)\n"
                    "  //frequency 4.0  // increases # of color \"spokes\"\n"
             )
    )
  )
)

(defun pov-im-ripples ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// especially suited for normal {}\n"
                    "  ripples 0.5    // make surface ripply (0...1 or more)\n"
                    "  //ripples      // for use with pigment{}\n"
                    "\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-slope ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment\n"
                    "// this pattern can only be used when applied to an object\n"
                    "// value is influenced by the slope of the object\n"
                    "  slope\n"
                    "    y,              // direction vector for slope\n"
                    "    //y,            // optional: dependency from altitude\n"
                    "    //<0.5,1.0>,    // optional: range of slope values\n"
                    "    //<0.75,1.0>    // optional: range of altitude values\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-spherical ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// originally designed for use with media density\n"
                    "  spherical     // spherical gradient around origin\n"
             )
    )
  )
)

(defun pov-im-spiral1 ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  spiral1 5,         // spiral that winds around the y-axis\n"
                    "  //spiral1 5, 0.5   // for use with normal{} (0...1 or more)\n"
             )
    )
  )
)

(defun pov-im-spiral2 ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "  spiral2 5,         // double spiral that winds around the y-axis\n"
                    "  //spiral2 5, 0.5   // for use with normal{} (0...1 or more)\n"
             )
    )
  )
)

(defun pov-im-spotted ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// spotted is identical to bozo\n"
                    "  spotted        // create a spotted texture\n"
                    "  //spotted 0.5  // for use with normal{} (0...1 or more)\n"
             )
    )
  )
)

(defun pov-im-waves ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// especially suited for normal {}\n"
                    "  waves 0.5   // create a wavy surface (0...1 or more)\n"
                    "  //waves     // for use with pigment{}\n"
             )
    )
  )
)

(defun pov-im-wood ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// create a concentric cylindrical wood-like texture,\n"
                    "// grain along Z axis\n"
                    "  wood\n"
                    "  //wood 0.5         // for use with normal{} (0...1 or more)\n"
                    "  //turbulence 0.1   // should have some!\n"
             )
    )
  )
)

(defun pov-im-wrinkles ()
  (interactive
    (insert (concat
                    "// pattern for use in texture/pigment/normal/density\n"
                    "// especially suited for normal {}\n"
                    "  wrinkles 0.5   // create a wrinkly surface (0...1 or more)\n"
                    "  //wrinkles      // for use with pigment{}\n"
             )
    )
  )
)

(defun pov-im-10---waveform-attribute ()
  (interactive
    (insert (concat
                    "// you can specify a waveform if patterns are used with blend maps.\n"
                    "// Especially useful for normals\n"
                    "// Note most patterns have default waveform ramp_wave\n"
                    "  ramp_wave         // linear raising from 0 to 1 and returning abrupt return\n"
                    "  //triangle_wave   // linear upwards and downwards\n"
                    "  //sine_wave       // sine waveform\n"
                    "  //scallop_wave    // absolute value of sine wave\n"
                    "  //cubic_wave      // cubic curve from 0.0 to 1.0\n"
                    "  //poly_wave 0.5   // exponential function\n"
             )
    )
  )
)

(defun pov-im-20---frequency-attribute ()
  (interactive
    (insert (concat
                    "// pigment/normal/density attribute\n"
                    "// oscillation frequency of radial, ripples, waves, etc\n"
                    "  frequency 2.0\n"
             )
    )
  )
)

(defun pov-im-30---phase-attribute ()
  (interactive
    (insert (concat
                    "// pigment/normal/density attribute\n"
                    "// change the phase of ripples, etc.\n"
                    "// Animation note:  As phase DECREASES, ripples move OUTWARD.\n"
                    "  phase 0.5\n"
             )
    )
  )
)

(defun pov-im-40---turbulence-attribute ()
  (interactive
    (insert (concat
                    "// pattern modifier\n"
                    "// Takes a float or vector for selective turbulence in different directions\n"
                    "  turbulence <0, 1, 0.5> // how much to 'stir up' the texture (0...1 or more)\n"
                    "  // (---additional parameters influencing style of turbulence---)\n"
                    "  lambda 2.0      // (1.0...5.0) (smooth...random) [2.0]\n"
                    "  omega 0.5       // (0.0...2.0) (crinkly...fuzzy) [0.5]\n"
                    "  octaves 6       // number of \"folds\" in turbulence (1...10) [6]\n"
             )
    )
  )
)

(defun pov-im-50---turbulence-warp ()
  (interactive
    (insert (concat
                    "// pattern modifier\n"
                    "  // (---turbulence warp---)\n"
                    "  // turbulates similar to 'turbulence'\n"
                    "  warp {\n"
                    "    turbulence 0.4*x     // amount of turbulence\n"
                    "    //octaves 2          // optional turbulence modifiers\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-60---black_hole-warp ()
  (interactive
    (insert (concat
                    "// pattern modifier\n"
                    "  // (---black_hole warp---)\n"
                    "  // local deformation of the pattern\n"
                    "  warp {\n"
                    "    black_hole <0.5, 0.5, 0>, 0.25   // position and radius\n"
                    "    //strength 3                     // strength of deformation\n"
                    "    //inverse                        // invert effect\n"
                    "    //repeat <1.5, 1, 0>             // repeat black hole\n"
                    "    //turbulence <0.1, 0.4, 0>       // turbulate repeat position\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-70---repeat-warp ()
  (interactive
    (insert (concat
                    "// pattern modifier\n"
                    "  // (---repeat warp---)\n"
                    "  // repeats a section of the pattern\n"
                    "  warp {\n"
                    "    repeat x*2           // repeat direction\n"
                    "    //flip y             // flip when repeating\n"
                    "    //offset z*0.25      // translate when repeating\n"
                    "  }\n"
             )
    )
  )
)

(defun pov-im-80---mapping-warps ()
  (interactive
    (insert (concat
                    "// pattern modifier\n"
                    "  // (---cylindrical warp---)\n"
                    "  warp {\n"
                    "    cylindrical\n"
                    "    orientation z    // mapping orientation\n"
                    "    //dist_exp 0.0   // optional [0.0]\n"
                    "  }\n"
                    "  // (---spherical warp---)\n"
                    "  warp {\n"
                    "    spherical\n"
                    "    orientation z    // mapping orientation\n"
                    "    //dist_exp 0.0   // optional [0.0]\n"
                    "  }\n"
                    "  // (---toroidal warp---)\n"
                    "  warp {\n"
                    "    toroidal\n"
                    "    orientation z    // mapping orientation\n"
                    "    //dist_exp 0.0   // optional [0.0]\n"
                    "    major_radius 1   // major radius of torus\n"
                    "  }\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-90---noise_generator ()
  (interactive
    (insert (concat
                    "// there are 3 available noise generators\n"
                    "// you can select one by adding this as a pattern modifier\n"
                    "  //noise_generator 1 // old 3.1g plateaud noise\n"
                    "  //noise_generator 2 // fixed version of 3.1 noise\n"
                    "  noise_generator 3   // New 3.5 perlin noise\n"
                    "\n"
             )
    )
  )
)

(defun pov-im-matrix ()
  (interactive
    (insert (concat
                    "// Matrix is an object transformation that does rotation about the Y axis,\n"
                    "// shear along the Y axis, and translation along the Y axis\n"
                    "matrix\n"
                    "<\n"
                    "  0.886, 0.5, 0.5,       // the first 3 lines form a rotation matrix\n"
                    "  0,     1,   0,         // since it is not orthogonal, shearing occurs\n"
                    "  0.5,   0,  -0.886,\n"
                    "  0,     1.5, 0          // the last 3 values contain the translation\n"
                    ">\n"
             )
    )
  )
)

(defun pov-im-rotate ()
  (interactive
    (insert (concat
                    "// rotate shape or texture around <0 0 0>, relative to current orientation\n"
                    "// example: rotate 30*x\n"
                    "rotate <45, 45, 0> // <dX, dY, dZ> (in degrees)\n"
             )
    )
  )
)

(defun pov-im-scale ()
  (interactive
    (insert (concat
                    "// \"resize\" an object or texture relative to its current size\n"
                    "// example: scale 2.0\n"
                    "scale <1.0,  2.0,  1.0> // <dX, dY, dZ>\n"
             )
    )
  )
)

(defun pov-im-transform ()
  (interactive
    (insert (concat
                    "// object/texture/etc. transformation\n"
                    "// do a predeclared translate/rotate/scale combo\n"
                    "// e.g.:\n"
                    "#declare Leaf_Orientation =\n"
                    "transform {\n"
                    "  scale 4.0\n"
                    "  rotate 30*z\n"
                    "  translate 2*y\n"
                    "}\n"
                    "\n"
                    "transform {\n"
                    "  Leaf_Orientation    // transform an object\n"
                    "  //scale 0.5\n"
                    "  //rotate 45*x\n"
                    "  //transform {...}   // optional further transforms\n"
                    "}\n"
             )
    )
  )
)

(defun pov-im-translate ()
  (interactive
    (insert (concat
                    "// Move object or texture relative to current position\n"
                    "// example: translate -5.0*z\n"
                    "translate <2.0, 4.0, -1.0>  // <dX, dY, dZ>\n"
             )
    )
  )
)

;; End of insert menu
;; 231 Entries generated
